#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Settings window holding all settings tab"""
import gettext
import gi
import pkg_resources
from .voice_settings import VoiceSettingsWindow
from .text_settings import TextSettingsWindow
from .general_settings import GeneralSettingsWindow
from .about_settings import AboutSettingsWindow
from .notification_settings import NotificationSettingsWindow
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk  # nopep8

t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'))
_ = t.gettext


class MainSettingsWindow(Gtk.Window):
    """Settings window holding all settings tab"""

    def __init__(self, discover):
        Gtk.Window.__init__(self)

        self.connect("destroy", self.close_window)
        self.connect("delete-event", self.close_window)

        self.discover = discover
        self.text_overlay = discover.text_overlay
        self.voice_overlay = discover.voice_overlay
        self.notification_overlay = discover.notification_overlay
        self.set_title(_("Discover Overlay Configuration"))
        self.set_icon_name("discover-overlay")
        self.set_default_size(280, 180)

        # Create
        notebook = Gtk.Notebook()

        if discover.steamos:
            notebook.set_tab_pos(Gtk.PositionType.LEFT)
            self.set_default_size(1280, 800)

        self.about_settings = AboutSettingsWindow(self.discover)
        about_label = Gtk.Label.new(_("Overview"))
        notebook.append_page(self.about_settings)
        notebook.set_tab_label(self.about_settings, about_label)

        self.voice_settings = VoiceSettingsWindow(self.voice_overlay, discover)
        voice_label = Gtk.Label.new(_("Voice"))
        notebook.append_page(self.voice_settings)
        notebook.set_tab_label(self.voice_settings, voice_label)

        self.text_settings = TextSettingsWindow(self.text_overlay, discover)
        text_label = Gtk.Label.new(_("Text"))
        notebook.append_page(self.text_settings)
        notebook.set_tab_label(self.text_settings, text_label)

        self.notification_settings = NotificationSettingsWindow(
            self.notification_overlay, discover)
        notification_label = Gtk.Label.new(_("Notifications"))
        notebook.append_page(self.notification_settings)
        notebook.set_tab_label(self.notification_settings, notification_label)

        self.core_settings = GeneralSettingsWindow(self.discover)
        core_label = Gtk.Label.new(_("Core"))
        notebook.append_page(self.core_settings)
        notebook.set_tab_label(self.core_settings, core_label)
        self.add(notebook)
        self.notebook = notebook

    def close_window(self, widget=None, event=None):
        """
        Hide the settings window for use at a later date
        """
        self.text_settings.close_window(widget, event)
        self.voice_settings.close_window(widget, event)
        self.core_settings.close_window(widget, event)
        self.hide()
        return True

    def present_settings(self):
        """
        Show the settings window
        """
        self.about_settings.present_settings()
        self.voice_settings.present_settings()
        self.text_settings.present_settings()
        self.notification_settings.present_settings()
        self.core_settings.present_settings()
        self.notebook.show()
        self.show()



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
Overlay parent class. Helpful if we need more overlay
types without copy-and-pasting too much code
"""
import sys
import logging
import gi
import cairo
import Xlib
from Xlib.display import Display
from Xlib import X, Xatom
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Gdk  # nopep8
try:
    gi.require_version('GtkLayerShell', '0.1')
    from gi.repository import GtkLayerShell
except (ImportError, ValueError):
    pass

log = logging.getLogger(__name__)

class OverlayWindow(Gtk.Window):
    """
    Overlay parent class. Helpful if we need more overlay
    types without copy-and-pasting too much code
    """

    def detect_type(self):
        """
        Helper function to determine if Wayland is being used and return the Window type needed
        """
        window = Gtk.Window()
        screen = window.get_screen()
        screen_type = "%s" % (screen)
        self.is_wayland = False
        if "Wayland" in screen_type:
            self.is_wayland = True
            return Gtk.WindowType.TOPLEVEL
        return Gtk.WindowType.POPUP

    def __init__(self, discover, piggyback=None):
        Gtk.Window.__init__(self, type=self.detect_type())
        self.discover = discover
        screen = self.get_screen()
        self.compositing = False
        self.text_font = None
        self.text_size = None
        self.pos_x = None
        self.pos_y = None
        self.width = None
        self.height = None
        self.needsredraw = True
        self.hidden = False
        self.enabled = False

        self.set_size_request(50, 50)
        self.connect('draw', self.overlay_draw)
        # Set RGBA
        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        if not self.get_display().supports_input_shapes():
            log.info(
                "Input shapes not available. Quitting")
            sys.exit(1)
        if visual:
            # Set the visual even if we can't use it right now
            self.set_visual(visual)
        if screen.is_composited():
            self.compositing = True

        self.set_app_paintable(True)
        self.set_untouchable()
        self.set_skip_pager_hint(True)
        self.set_skip_taskbar_hint(True)
        self.set_keep_above(True)
        self.set_decorated(True)
        self.set_accept_focus(False)
        self.set_wayland_state()
        if not piggyback:
            self.show_all()
            if discover.steamos:
                display = Display()
                atom = display.intern_atom("GAMESCOPE_EXTERNAL_OVERLAY")
                opaq = display.intern_atom("_NET_WM_WINDOW_OPACITY")

                topw = display.create_resource_object(
                    "window", self.get_toplevel().get_window().get_xid())

                topw.change_property(atom,
                                     Xatom.CARDINAL, 32,
                                     [1], X.PropModeReplace)
                # Keep for reference, but appears to be unnecessary
                # topw.change_property(opaq,
                #                     Xatom.CARDINAL,16,
                #                     [0xffff], X.PropModeReplace)

                log.info("Setting STEAM_EXTERNAL_OVERLAY")
                display.sync()
        self.monitor = 0
        self.align_right = True
        self.align_vert = 1
        self.floating = False
        self.force_xshape = False
        self.context = None
        self.autohide = False
        self.piggyback = None
        self.piggyback_parent = None
        if piggyback:
            self.set_piggyback(piggyback)

    def set_wayland_state(self):
        """
        If wayland is in use then attempt to set up a GtkLayerShell
        I have no idea how this should register a fail for Weston/Gnome
        """
        if self.is_wayland:
            if not GtkLayerShell.is_supported():
                log.info(
                    "GTK Layer Shell is not supported on this wayland compositor")
                log.info("Currently not possible: Gnome, Weston")
                sys.exit(0)
            GtkLayerShell.init_for_window(self)
            GtkLayerShell.set_layer(self, GtkLayerShell.Layer.OVERLAY)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)

    def set_piggyback(self, other_overlay):
        other_overlay.piggyback = self
        self.piggyback_parent = other_overlay

    def overlay_draw(self, _w, context, data=None):
        """
        Draw overlay
        """

    def set_font(self, font):
        """
        Set the font used by the overlay
        """
        self.text_font = font
        self.needsredraw = True

    def set_floating(self, floating, pos_x, pos_y, width, height):
        """
        Set if the window is floating and what dimensions to use
        """

        self.floating = floating
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.width = width
        self.height = height
        self.force_location()

    def set_untouchable(self):
        """
        Create a custom input shape and tell it that all of the window is a cut-out
        This allows us to have a window above everything but that never gets clicked on
        """
        (width, height) = self.get_size()
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        surface_ctx = cairo.Context(surface)
        surface_ctx.set_source_rgba(0.0, 0.0, 0.0, 0.0)
        surface_ctx.set_operator(cairo.OPERATOR_SOURCE)
        surface_ctx.paint()
        reg = Gdk.cairo_region_create_from_surface(surface)
        self.input_shape_combine_region(reg)

    def unset_shape(self):
        """
        Remove XShape (not input shape)
        """
        self.get_window().shape_combine_region(None, 0, 0)

    def force_location(self):
        """
        On X11 enforce the location and sane defaults
        On Wayland just store for later
        On Gamescope enforce size of display but only if it's the primary overlay
        """
        if self.discover.steamos and not self.piggyback_parent:
            display = Gdk.Display.get_default()
            if "get_monitor" in dir(display):
                monitor = display.get_monitor(self.monitor)
                geometry = monitor.get_geometry()
                scale_factor = monitor.get_scale_factor()
                width = geometry.width
                height = geometry.height
                self.resize(width, height)
                self.needsredraw = True
                return
        if not self.is_wayland:
            self.set_decorated(False)
            self.set_keep_above(True)
            display = Gdk.Display.get_default()
            if "get_monitor" in dir(display):
                monitor = display.get_monitor(self.monitor)
                geometry = monitor.get_geometry()
                scale_factor = monitor.get_scale_factor()
                if not self.floating:
                    width = geometry.width
                    height = geometry.height
                    pos_x = geometry.x
                    pos_y = geometry.y
                    self.resize(width, height)
                    self.move(pos_x, pos_y)
                else:
                    self.move(self.pos_x, self.pos_y)
                    self.resize(self.width, self.height)
            else:
                if self.floating:
                    self.move(self.pos_x, self.pos_y)
                    self.resize(self.width, self.height)

        if not self.floating:
            (width, height) = self.get_size()
            self.width = width
            self.height = height
        self.needsredraw = True

    def redraw(self):
        """
        Request a redraw.
        If we're using XShape (optionally or forcibly) then render the image into the shape
        so that we only cut out clear sections
        """
        self.needsredraw = False
        gdkwin = self.get_window()
        if self.piggyback_parent:
            self.piggyback_parent.redraw()
            return
        if not self.floating:
            (width, height) = self.get_size()
            self.width = width
            self.height = height
        if gdkwin:
            if not self.compositing or self.force_xshape:
                (width, height) = self.get_size()
                surface = cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, width, height)
                surface_ctx = cairo.Context(surface)
                self.overlay_draw(None, surface_ctx)
                reg = Gdk.cairo_region_create_from_surface(surface)
                gdkwin.shape_combine_region(reg, 0, 0)
            else:
                gdkwin.shape_combine_region(None, 0, 0)
        self.queue_draw()

    def set_hidden(self, hidden):
        self.hidden = hidden
        self.set_enabled(self.enabled)

    def set_monitor(self, idx=None, mon=None):
        """
        Set the monitor this overlay should display on.
        """
        self.monitor = idx
        if self.is_wayland:
            if mon:
                GtkLayerShell.set_monitor(self, mon)
        self.force_location()
        self.needsredraw = True

    def set_align_x(self, align_right):
        """
        Set the alignment (True for right, False for left)
        """
        self.align_right = align_right
        self.force_location()
        self.needsredraw = True

    def set_align_y(self, align_vert):
        """
        Set the veritcal alignment
        """
        self.align_vert = align_vert
        self.force_location()
        self.needsredraw = True

    def col(self, col, alpha=1.0):
        """
        Convenience function to set the cairo context next colour
        """
        self.context.set_source_rgba(col[0], col[1], col[2], col[3] * alpha)

    def set_force_xshape(self, force):
        """
        Set if XShape should be forced
        """
        self.force_xshape = force

        if self.is_wayland:
            # Wayland and XShape are a bad idea unless you're a fan on artifacts
            self.force_xshape = False

    def set_enabled(self, enabled):
        """
        Set if this overlay should be visible
        """
        self.enabled = enabled
        if enabled and not self.hidden and not self.piggyback_parent:
            self.show_all()
            self.set_untouchable()
        else:
            self.hide()

    def set_hide_on_mouseover(self, hide):
        """
        Set Mouseover hide
        """
        self.autohide = hide

    def set_task(self, visible):
        self.set_skip_pager_hint(not visible)
        self.set_skip_taskbar_hint(not visible)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Functions & Classes to assist image loading."""
import urllib
import threading
import logging
import gi
import requests
import cairo
import PIL
import PIL.Image as Image
import os
import io
gi.require_version('GdkPixbuf', '2.0')
# pylint: disable=wrong-import-position
from gi.repository import Gio, GdkPixbuf  # nopep8

log = logging.getLogger(__name__)

class SurfaceGetter():
    """Download and decode image using PIL and store as a cairo surface"""

    def __init__(self, func, url, identifier, size):
        self.func = func
        self.identifier = identifier
        self.url = url
        self.size = size

    def get_url(self):
        """Downloads and decodes"""
        try:
            resp = requests.get(
                self.url, stream=True, headers={
                    'Referer': 'https://streamkit.discord.com/overlay/voice',
                    'User-Agent': 'Mozilla/5.0'
                }
            )
            raw = resp.raw
            image = Image.open(raw)
            surface = from_pil(image)

            self.func(self.identifier, surface)
        except requests.HTTPError:
            log.error("Unable to open %s", self.url)
        except requests.TooManyRedirects:
            log.error("Unable to open %s - Too many redirects", self.url)
        except requests.Timeout:
            log.error("Unable to open %s - Timeout", self.url)
        except requests.ConnectionError:
            log.error("Unable to open %s - Connection error", self.url)
        except ValueError:
            log.error("Unable to read %s", self.url)
        except TypeError:
            log.error("Unable to read %s", self.url)
        except PIL.UnidentifiedImageError:
            log.error("Unknown image type")

    def get_file(self):
        locations = [os.path.expanduser('~/.local/'), '/usr/', '/app']
        for prefix in locations:
            mixpath = os.path.join(prefix, self.url)
            image = None
            try:
                image = Image.open(mixpath)
            except ValueError:
                log.error("Value Erorr - Unable to read %s", mixpath)
            except TypeError:
                log.error("Type Error - Unable to read %s", mixpath)
            except PIL.UnidentifiedImageError:
                log.error("Unknown image type: %s", mixpath)
            except FileNotFoundError:
                log.error("File not found: %s", mixpath)
            if image:
                surface = from_pil(image)
                if surface:
                    self.func(self.identifier, surface)
                    return


def from_pil(image, alpha=1.0):
    """
    :param im: Pillow Image
    :param alpha: 0..1 alpha to add to non-alpha images
    :param format: Pixel format for output surface
    """
    if 'A' not in image.getbands():
        image.putalpha(int(alpha * 256.))
    arr = bytearray(image.tobytes('raw', 'BGRa'))
    surface = cairo.ImageSurface.create_for_data(
        arr, cairo.FORMAT_ARGB32, image.width, image.height)
    return surface

def get_surface(func, identifier, ava, size):
    """Download to cairo surface"""
    image_getter = SurfaceGetter(func, identifier, ava, size)
    if identifier.startswith('http'):
        thread = threading.Thread(target=image_getter.get_url, args=())
        thread.start()
    else:
        thread = threading.Thread(target=image_getter.get_file, args=())
        thread.start()


def make_surface_from_raw(raw, size):
    """Create surface from raw notification data"""
    width = raw[0]
    height = raw[1]
    rowstride = raw[2]
    hasalpha = raw[3]
    bitspersample = raw[4]
    channels = raw[5]
    image_raw_dbus = raw[6]
    image_raw = bytes(image_raw_dbus)
    image = None
    if hasalpha:
        image = Image.frombytes('RGBA', [width, height], image_raw, 'raw')
    else:
        image = Image.frombytes('RGB', [width, height], image_raw, 'raw')
    surface = from_pil(image)
    return surface


def get_aspected_size(img, width, height, anchor=0, hanchor=0):
    """Get dimensions of image keeping current aspect ratio"""
    pic_width = img.get_width()
    pic_height = img.get_height()
    if pic_height < 1 or height < 1:
        return (0, 0, 0, 0)
    img_aspect = pic_width / pic_height
    rect_aspect = width / height

    offset_y = 0
    offset_x = 0
    if img_aspect > rect_aspect:
        old_height = height
        height = width / img_aspect
        if anchor == 0:
            offset_y = offset_y + (old_height - height)
        if anchor == 1:
            offset_y = offset_y + ((old_height - height) / 2)
    elif img_aspect < rect_aspect:
        old_width = width
        width = height * img_aspect
        if hanchor == 2:
            offset_x = offset_x + (old_width - width)
        if hanchor == 1:
            offset_x = offset_x + ((old_width - width) / 2)
    return (offset_x, offset_y, width, height)


def draw_img_to_rect(img, ctx,
                     pos_x, pos_y,
                     width, height,
                     path=False, aspect=False,
                     anchor=0, hanchor=0):
    """Draw cairo surface onto context

    Path - only add the path do not fill : True/False
    Aspect - keep aspect ratio : True/False
    Anchor - with aspect : 0=left 1=middle 2=right
    HAnchor - with apect : 0=bottom 1=middle 2=top
    """

    ctx.save()
    offset_x = 0
    offset_y = 0
    if aspect:
        (offset_x, offset_y, width, height) = get_aspected_size(
            img, width, height, anchor=anchor, hanchor=hanchor)

    ctx.translate(pos_x + offset_x, pos_y + offset_y)
    ctx.scale(width, height)
    ctx.scale(1 / img.get_width(), 1 / img.get_height())
    ctx.set_source_surface(img, 0, 0)

    ctx.rectangle(0, 0, img.get_width(), img.get_height())
    if not path:
        ctx.fill()
    ctx.restore()
    return (width, height)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Main application class"""
import gettext
import os
import time
import sys
import re
import traceback
import logging
import pkg_resources
import gi
import pidfile
from .settings_window import MainSettingsWindow
from .voice_overlay import VoiceOverlayWindow
from .text_overlay import TextOverlayWindow
from .notification_overlay import NotificationOverlayWindow
from .discord_connector import DiscordConnector

gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, GLib, Gio  # nopep8

try:
    from xdg.BaseDirectory import xdg_config_home
except ModuleNotFoundError:
    from xdg import XDG_CONFIG_HOME as xdg_config_home

log = logging.getLogger(__name__)
t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'))
_ = t.gettext


class Discover:
    """Main application class"""

    def __init__(self, rpc_file, debug_file, args):
        self.ind = None
        self.tray = None
        self.steamos = False
        self.connection = None
        self.show_settings_delay = False
        self.settings = None

        self.debug_file = debug_file

        self.do_args(args, True)
        if "GAMESCOPE_WAYLAND_DISPLAY" in os.environ:
            log.info(
                "GameScope session detected. Enabling steam and gamescope integration")
            self.steamos = True
            self.show_settings_delay = True
            settings = Gtk.Settings.get_default()
            if settings:
                settings.set_property(
                    "gtk-application-prefer-dark-theme", Gtk.true)

        self.create_gui()

        self.connection = DiscordConnector(
            self,
            self.settings.text_settings,
            self.settings.voice_settings,
            self.text_overlay,
            self.voice_overlay
        )

        self.settings.text_settings.add_connector(self.connection)
        self.connection.connect()
        GLib.timeout_add((1000 / 60), self.connection.do_read)
        GLib.timeout_add((1000 / 20), self.periodic_run)
        self.rpc_file = rpc_file
        rpc_file = Gio.File.new_for_path(rpc_file)
        monitor = rpc_file.monitor_file(0, None)
        monitor.connect("changed", self.rpc_changed)

        Gtk.main()

    def periodic_run(self, data=None):
        if self.voice_overlay.needsredraw:
            self.voice_overlay.redraw()

        if self.text_overlay:
            self.text_overlay.tick()
            if self.text_overlay.needsredraw:
                self.text_overlay.redraw()

        if self.notification_overlay and self.notification_overlay.enabled:
            self.notification_overlay.tick()
            if self.notification_overlay.needsredraw:
                self.notification_overlay.redraw()
        return True

    def do_args(self, data, normal_close):
        """
        Read in arg list from command or RPC and act accordingly
        """
        if "--help" in data or "-h" in data:
            print("%s: discover-overlay [OPTIONS]... " % (_("Usage")))
            print(_("Show an X11 or wlroots overlay with information"))
            print(_("from Discord client"))
            print("")
            print("  -c, --configure        ", _("Open configuration window"))
            print("  -x, --close            ",
                  _("Close currently running instance"))
            print("  -v, --debug            ",
                  _("Verbose output for aid in debugging"))
            print("  -h, --help             ", _("This screen"))
            print("      --hide             ", _("Hide overlay"))
            print("      --show             ", _("Show overlay"))
            print("      --rpc              ",
                  _("Send command, not start new instance. Only needed if running in flatpak"))
            print("      --mute             ", _("Set own user to mute"))
            print("      --unmute           ", _("Set unmuted"))
            print("      --deaf             ", _("Set own user to deafened"))
            print("      --undeaf           ", _("Unset user deafened state"))
            print("      --moveto=XX        ",
                  _("Move the user into voice room, by Room ID"))
            print("")
            print(_("For gamescope compatibility ensure ENV has 'GDK_BACKEND=x11'"))
            if normal_close:
                sys.exit(0)
        if "--configure" in data or "-c" in data:
            if self.settings:
                self.show_settings()
            else:
                self.show_settings_delay = True
        if "--close" in data or "-x" in data:
            sys.exit(0)
        if "--steamos" in data or "-s" in data:
            self.steamos = True
        if "--hide" in data:
            if self.voice_overlay:
                self.voice_overlay.set_hidden(True)
            if self.text_overlay:
                self.text_overlay.set_hidden(True)
        if "--show" in data:
            if self.voice_overlay:
                self.voice_overlay.set_hidden(False)
            if self.text_overlay:
                self.text_overlay.set_hidden(False)
        if "--mute" in data:
            if self.connection:
                self.connection.set_mute(True)
        if "--unmute" in data:
            if self.connection:
                self.connection.set_mute(False)
        if "--deaf" in data:
            if self.connection:
                self.connection.set_deaf(True)
        if "--undeaf" in data:
            if self.connection:
                self.connection.set_deaf(False)
        pattern = re.compile("--moveto=([0-9]+)")
        if any((match := pattern.match(x)) for x in data):
            if self.connection:
                self.connection.change_voice_room(match.group(1))

    def rpc_changed(self, _a=None, _b=None, _c=None, _d=None):
        """
        Called when the RPC file has been altered
        """
        with open(self.rpc_file, "r") as tfile:
            data = tfile.readlines()
            if len(data) >= 1:
                self.do_args(data[0].split(" "), False)

    def create_gui(self):
        """
        Create Systray & associated menu, overlays & settings windows
        """
        self.voice_overlay = VoiceOverlayWindow(self)

        if self.steamos:
            self.text_overlay = TextOverlayWindow(self, self.voice_overlay)
            self.notification_overlay = NotificationOverlayWindow(
                self, self.text_overlay)
        else:
            self.text_overlay = TextOverlayWindow(self)
            self.notification_overlay = NotificationOverlayWindow(self)
        self.menu = self.make_menu()
        self.make_sys_tray_icon(self.menu)
        self.settings = MainSettingsWindow(self)

        if self.steamos:
            # Larger fonts needed
            css = Gtk.CssProvider.new()
            css.load_from_data(bytes("* { font-size:20px; }", "utf-8"))
            self.settings.get_style_context().add_provider(
                css, Gtk.STYLE_PROVIDER_PRIORITY_USER)

    def make_sys_tray_icon(self, menu):
        """
        Attempt to create an AppIndicator icon, failing that attempt to make
        a systemtray icon
        """
        if self.steamos:
            return
        try:
            gi.require_version('AppIndicator3', '0.1')
            # pylint: disable=import-outside-toplevel
            from gi.repository import AppIndicator3
            self.ind = AppIndicator3.Indicator.new(
                "discover_overlay",
                "discover-overlay-tray",
                AppIndicator3.IndicatorCategory.APPLICATION_STATUS)
            # Hide for now since we don't know if it should be shown yet
            self.ind.set_status(AppIndicator3.IndicatorStatus.PASSIVE)
            self.ind.set_menu(menu)
        except (ImportError, ValueError) as exception:
            # Create System Tray
            log.info("Falling back to Systray : %s", exception)
            self.tray = Gtk.StatusIcon.new_from_icon_name(
                "discover-overlay-tray")
            self.tray.connect('popup-menu', self.show_menu)
            # Hide for now since we don't know if it should be shown yet
            self.tray.set_visible(False)

    def make_menu(self):
        """
        Create System Menu
        """
        menu = Gtk.Menu()
        settings_opt = Gtk.MenuItem.new_with_label(_("Settings"))
        close_opt = Gtk.MenuItem.new_with_label(_("Close"))

        menu.append(settings_opt)
        menu.append(close_opt)

        settings_opt.connect("activate", self.show_settings)
        close_opt.connect("activate", self.close)
        menu.show_all()
        return menu

    def show_menu(self, obj, button, time):
        """
        Show menu when System Tray icon is clicked
        """
        self.menu.show_all()
        self.menu.popup(
            None, None, Gtk.StatusIcon.position_menu, obj, button, time)

    def show_settings(self, _obj=None, _data=None):
        """
        Show settings window
        """
        self.settings.present_settings()

    def hide_settings(self, _obj=None, _data=None):
        """
        Hide settings window
        """
        self.settings.close_window()

    def close(self, _a=None, _b=None, _c=None):
        """
        End of the program
        """
        Gtk.main_quit()

    def set_force_xshape(self, force):
        """
        Set if XShape should be forced
        """
        self.voice_overlay.set_force_xshape(force)
        if self.text_overlay:
            self.text_overlay.set_force_xshape(force)

    def set_show_task(self, visible):
        if self.voice_overlay:
            self.voice_overlay.set_task(visible)
        if self.text_overlay:
            self.text_overlay.set_task(visible)

    def set_sys_tray_icon_visible(self, visible):
        """
        Sets whether the tray icon is visible
        """
        if self.ind is not None:
            # pylint: disable=import-outside-toplevel
            from gi.repository import AppIndicator3
            self.ind.set_status(
                AppIndicator3.IndicatorStatus.ACTIVE if visible else AppIndicator3.IndicatorStatus.PASSIVE)
        elif self.tray is not None:
            self.tray.set_visible(visible)


def entrypoint():
    """
    Entry Point.

    Check for PID & RPC.

    If an overlay is already running then pass the args along and close

    Otherwise start up the overlay!
    """

    config_dir = os.path.join(xdg_config_home, "discover_overlay")
    os.makedirs(config_dir, exist_ok=True)
    line = ""
    for arg in sys.argv[1:]:
        line = "%s %s" % (line, arg)

    pid_file = os.path.join(config_dir, "discover_overlay.pid")
    rpc_file = os.path.join(config_dir, "discover_overlay.rpc")
    debug_file = os.path.join(config_dir, "output.txt")
    logging.getLogger().setLevel(logging.INFO)
    FORMAT = "%(levelname)s - %(name)s - %(message)s"
    if "--debug" in sys.argv or "-v" in sys.argv:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.basicConfig(filename=debug_file, format=FORMAT)
    else:
        logging.basicConfig(format=FORMAT)
    log = logging.getLogger(__name__)
    log.info("Starting Discover Overlay: %s",
             pkg_resources.get_distribution('discover_overlay').version)

    # Flatpak compat mode
    try:
        if "container" in os.environ and os.environ["container"] == "flatpak":
            if "--rpc" in sys.argv:
                with open(rpc_file, "w") as tfile:
                    tfile.write(line)
                    log.warning("Sent RPC command")
            else:
                log.info("Flatpak compat mode started")
                with open(rpc_file, "w") as tfile:
                    tfile.write("--close")
                Discover(rpc_file, debug_file, sys.argv[1:])
            return

        # Normal usage

        try:
            with pidfile.PIDFile(pid_file):
                Discover(rpc_file, debug_file, sys.argv[1:])
        except pidfile.AlreadyRunningError:
            log.warning("Discover overlay is currently running")

            with open(rpc_file, "w") as tfile:
                tfile.write(line)
                log.warning("Sent RPC command")
    except Exception as ex:
        log.error(ex)
        log.error(traceback.format_exc())
        sys.exit(1)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Text setting tab on settings window"""
import gettext
import json
import logging
import pkg_resources
from configparser import ConfigParser
import gi
from .settings import SettingsWindow

gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Gdk  # nopep8


GUILD_DEFAULT_VALUE = "0"
log = logging.getLogger(__name__)
t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'))
_ = t.gettext


class TextSettingsWindow(SettingsWindow):
    """Text setting tab on settings window"""

    def __init__(self, overlay, discover):
        SettingsWindow.__init__(self, discover)
        self.overlay = overlay

        self.placement_window = None
        self.init_config()
        self.list_channels_keys = []
        self.list_channels = {}
        self.list_guilds_keys = []
        self.list_guilds = {}
        self.ignore_channel_change = False
        self.ignore_guild_change = False
        self.channel_lookup = None
        self.channel_model = None
        self.connector = None
        self.guild_lookup = None
        self.guild_model = None
        self.guild_widget = None
        self.align_x = None
        self.align_y = None
        self.monitor = None
        self.floating = None
        self.channel = None
        self.guild = None
        self.font = None
        self.bg_col = None
        self.fg_col = None
        self.popup_style = None
        self.text_time = None
        self.show_attach = None
        self.line_limit = None
        self.enabled = None

        self.set_size_request(400, 200)
        self.connect("destroy", self.close_window)
        self.connect("delete-event", self.close_window)
        if overlay:
            self.init_config()
            self.create_gui()

    def update_channel_model(self):
        """
        Update the Channel selector.

        Populate with all channels from guild.
        Leave empty if guild is unselected
        """
        # Potentially organize channels by their group/parent_id
        # https://discord.com/developers/docs/resources/channel#channel-object-channel-structure
        c_model = Gtk.ListStore(str, bool)
        self.channel_lookup = []

        # If a guild is specified, populate channel list with every channel from *just that guild*
        if self.guild != GUILD_DEFAULT_VALUE:
            for position in self.list_channels_keys:
                chan = self.list_channels[position]
                channel_key = chan["id"]
                if chan['guild_id'] == self.guild:
                    c_model.append([chan["name"], True])
                    self.channel_lookup.append(channel_key)

        self.channel_widget.set_model(c_model)
        self.channel_model = c_model

        idx = 0
        for channel in self.channel_lookup:
            if channel == self.channel:
                self.ignore_channel_change = True
                self.channel_widget.set_active(idx)
                self.ignore_channel_change = False
                break
            idx += 1
        self.channel_widget.set_sensitive(True)

    def add_connector(self, conn):
        """
        Add the discord_connector reference

        If the user has previously selected a text channel then tell it to subscribe
        """
        self.connector = conn
        if self.channel:
            self.connector.set_text_channel(self.channel)

    def present_settings(self):
        """
        Show contents of tab and update lists
        """
        if not self.overlay:
            return
        self.show_all()
        if not self.floating:
            self.align_x_widget.show()
            self.align_y_widget.show()
            self.align_monitor_widget.show()
            self.align_placement_widget.hide()
        else:
            self.align_x_widget.hide()
            self.align_y_widget.hide()
            self.align_monitor_widget.show()
            self.align_placement_widget.show()

        if self.popup_style:
            self.text_time_widget.show()
            self.text_time_label_widget.show()
        else:
            self.text_time_widget.hide()
            self.text_time_label_widget.hide()
        self.update_guild_model()

    def update_guild_model(self):

        g_model = Gtk.ListStore(str, bool)
        self.guild_lookup = []

        for guild in self.guild_list():
            guild_id, guild_name = guild
            self.guild_lookup.append(guild_id)
            g_model.append([guild_name, True])

        self.guild_widget.set_model(g_model)
        self.guild_model = g_model
        self.update_channel_model()

        idxg = 0
        for guild_id in self.guild_lookup:
            if guild_id == self.guild:
                self.ignore_guild_change = True
                self.guild_widget.set_active(idxg)
                self.ignore_guild_change = False
                break
            idxg += 1

        if self.guild is not None:
            self.refresh_channel_list()
        self.guild_widget.set_sensitive(True)

    def guild_list(self):
        """
        Return a list of all guilds
        """
        guilds = []
        done = []
        for guild in self.list_guilds.values():
            if not guild["id"] in done:
                done.append(guild["id"])
                guilds.append([guild["id"], guild["name"]])
        return guilds

    def set_channels(self, in_list):
        """
        Set the contents of list_channels
        """
        self.list_channels = in_list
        self.list_channels_keys = []
        for (key, _value) in enumerate(in_list):
            # Filter for only text channels
            # https://discord.com/developers/docs/resources/channel#channel-object-channel-types
            if in_list[key] is not None and in_list[key]["type"] == 0:
                self.list_channels_keys.append(key)
        self.list_channels_keys.sort()
        self.update_channel_model()

    def set_guilds(self, in_list):
        """
        Set the contents of list_guilds
        """
        self.guild_widget.set_sensitive(True)
        if self.list_guilds == in_list:
            return
        self.list_guilds = in_list
        self.list_guilds_keys = []
        for key in in_list.keys():
            self.list_guilds_keys.append(key)
        self.list_guilds_keys.sort()
        self.update_guild_model()

    def read_config(self):
        """
        Read in the 'text' section of the config
        """
        if not self.overlay:
            return
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        self.enabled = config.getboolean("text", "enabled", fallback=False)
        self.align_x = config.getboolean("text", "rightalign", fallback=True)
        self.align_y = config.getint("text", "topalign", fallback=2)
        self.monitor = config.get("text", "monitor", fallback="None")
        self.floating = config.getboolean("text", "floating", fallback=True)
        self.floating_x = config.getint("text", "floating_x", fallback=0)
        self.floating_y = config.getint("text", "floating_y", fallback=0)
        self.floating_w = config.getint("text", "floating_w", fallback=400)
        self.floating_h = config.getint("text", "floating_h", fallback=400)
        self.channel = config.get("text", "channel", fallback="0")
        self.guild = config.get("text", "guild", fallback=GUILD_DEFAULT_VALUE)
        self.font = config.get("text", "font", fallback=None)
        self.bg_col = json.loads(config.get(
            "text", "bg_col", fallback="[0.0,0.0,0.0,0.5]"))
        self.fg_col = json.loads(config.get(
            "text", "fg_col", fallback="[1.0,1.0,1.0,1.0]"))
        self.popup_style = config.getboolean(
            "text", "popup_style", fallback=False)
        self.text_time = config.getint("text", "text_time", fallback=30)
        self.show_attach = config.getboolean(
            "text", "show_attach", fallback=True)
        self.line_limit = config.getint("text", "line_limit", fallback=20)
        self.autohide = config.getboolean("text", "autohide", fallback=False)

        log.info(
            "Loading saved channel %s", self.channel)

        # Pass all of our config over to the overlay
        self.overlay.set_enabled(self.enabled)
        self.overlay.set_align_x(self.align_x)
        self.overlay.set_align_y(self.align_y)
        self.overlay.set_monitor(self.get_monitor_index(
            self.monitor), self.get_monitor_obj(self.monitor))
        self.overlay.set_floating(
            self.floating, self.floating_x, self.floating_y, self.floating_w, self.floating_h)
        self.overlay.set_bg(self.bg_col)
        self.overlay.set_fg(self.fg_col)
        self.overlay.set_popup_style(self.popup_style)
        self.overlay.set_text_time(self.text_time)
        self.overlay.set_show_attach(self.show_attach)
        self.overlay.set_line_limit(self.line_limit)
        self.overlay.set_hide_on_mouseover(self.autohide)

        if self.font:
            self.overlay.set_font(self.font)

    def save_config(self):
        """
        Save the current settings to the 'text' section of the config
        """
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        if not config.has_section("text"):
            config.add_section("text")

        config.set("text", "rightalign", "%d" % (int(self.align_x)))
        config.set("text", "topalign", "%d" % (self.align_y))
        config.set("text", "monitor", self.monitor)
        config.set("text", "enabled", "%d" % (int(self.enabled)))
        config.set("text", "floating", "%s" % (int(self.floating)))
        config.set("text", "floating_x", "%s" % (int(self.floating_x)))
        config.set("text", "floating_y", "%s" % (int(self.floating_y)))
        config.set("text", "floating_w", "%s" % (int(self.floating_w)))
        config.set("text", "floating_h", "%s" % (int(self.floating_h)))
        config.set("text", "channel", self.channel)
        config.set("text", "guild", self.guild)
        config.set("text", "bg_col", json.dumps(self.bg_col))
        config.set("text", "fg_col", json.dumps(self.fg_col))
        config.set("text", "popup_style", "%s" % (int(self.popup_style)))
        config.set("text", "text_time", "%s" % (int(self.text_time)))
        config.set("text", "line_limit", "%s" % (int(self.line_limit)))
        config.set("text", "show_attach", "%s" % (int(self.show_attach)))

        if self.font:
            config.set("text", "font", self.font)

        with open(self.config_file, 'w') as file:
            config.write(file)

    def create_gui(self):
        """
        Prepare the gui
        """
        box = Gtk.Grid()

        # Enabled
        enabled_label = Gtk.Label.new(_("Enable"))
        enabled = Gtk.CheckButton.new()
        enabled.set_active(self.enabled)
        enabled.connect("toggled", self.change_enabled)

        # Autohide
        autohide_label = Gtk.Label.new(_("Hide on mouseover"))
        autohide = Gtk.CheckButton.new()
        autohide.set_active(self.autohide)
        autohide.connect("toggled", self.change_hide_on_mouseover)

        # Popup Style
        popup_style_label = Gtk.Label.new(_("Popup Style"))
        popup_style = Gtk.CheckButton.new()
        popup_style.set_active(self.popup_style)
        popup_style.connect("toggled", self.change_popup_style)

        # Popup timer
        text_time_label = Gtk.Label.new(_("Popup timer"))
        text_time_adjustment = Gtk.Adjustment.new(
            self.text_time, 8, 9000, 1, 1, 8)
        text_time = Gtk.SpinButton.new(text_time_adjustment, 0, 0)
        text_time.connect("value-changed", self.change_text_time)

        # Font chooser
        font_label = Gtk.Label.new(_("Font"))
        font = Gtk.FontButton()
        if self.font:
            font.set_font(self.font)
        font.connect("font-set", self.change_font)

        # Colours
        bg_col_label = Gtk.Label.new(_("Background colour"))
        bg_col = Gtk.ColorButton.new_with_rgba(
            Gdk.RGBA(self.bg_col[0], self.bg_col[1], self.bg_col[2], self.bg_col[3]))
        fg_col_label = Gtk.Label.new(_("Text colour"))
        fg_col = Gtk.ColorButton.new_with_rgba(
            Gdk.RGBA(self.fg_col[0], self.fg_col[1], self.fg_col[2], self.fg_col[3]))
        bg_col.set_use_alpha(True)
        fg_col.set_use_alpha(True)
        bg_col.connect("color-set", self.change_bg)
        fg_col.connect("color-set", self.change_fg)

        # Monitor & Alignment
        align_label = Gtk.Label.new(_("Overlay Location"))

        align_type_box = Gtk.HBox()
        align_type_edge = Gtk.RadioButton.new_with_label(
            None, _("Anchor to edge"))
        align_type_floating = Gtk.RadioButton.new_with_label_from_widget(
            align_type_edge, _("Floating"))
        if self.floating:
            align_type_floating.set_active(True)
        align_type_box.add(align_type_edge)
        align_type_box.add(align_type_floating)

        monitor_store = Gtk.ListStore(str)
        display = Gdk.Display.get_default()
        if "get_n_monitors" in dir(display):
            for i in range(0, display.get_n_monitors()):
                monitor_store.append([display.get_monitor(i).get_model()])
        monitor = Gtk.ComboBox.new_with_model(monitor_store)
        monitor.set_active(self.get_monitor_index(self.monitor))
        monitor.connect("changed", self.change_monitor)
        renderer_text = Gtk.CellRendererText()
        monitor.pack_start(renderer_text, True)
        monitor.add_attribute(renderer_text, "text", 0)

        align_x_store = Gtk.ListStore(str)
        align_x_store.append([_("Left")])
        align_x_store.append([_("Right")])
        align_x = Gtk.ComboBox.new_with_model(align_x_store)
        align_x.set_active(True if self.align_x else False)
        align_x.connect("changed", self.change_align_x)
        renderer_text = Gtk.CellRendererText()
        align_x.pack_start(renderer_text, True)
        align_x.add_attribute(renderer_text, "text", 0)

        align_y_store = Gtk.ListStore(str)
        align_y_store.append([_("Top")])
        align_y_store.append([_("Middle")])
        align_y_store.append([_("Bottom")])
        align_y = Gtk.ComboBox.new_with_model(align_y_store)
        align_y.set_active(self.align_y)
        align_y.connect("changed", self.change_align_y)
        renderer_text = Gtk.CellRendererText()
        align_y.pack_start(renderer_text, True)
        align_y.add_attribute(renderer_text, "text", 0)

        align_placement_button = Gtk.Button.new_with_label(_("Place Window"))

        align_type_edge.connect("toggled", self.change_align_type_edge)
        align_type_floating.connect("toggled", self.change_align_type_floating)
        align_placement_button.connect("pressed", self.change_placement)

        channel_label = Gtk.Label.new(_("Channel"))
        channel = Gtk.ComboBox.new()

        channel.connect("changed", self.change_channel)
        renderer_text = Gtk.CellRendererText()
        channel.pack_start(renderer_text, True)
        channel.add_attribute(renderer_text, "text", 0)
        channel.add_attribute(renderer_text, 'sensitive', 1)

        channel_refresh = Gtk.Button.new_with_label(_("Refresh list"))
        channel_refresh.connect("pressed", self.refresh_channel_list)

        guild_label = Gtk.Label.new(_("Server"))
        guild = Gtk.ComboBox.new()

        guild.connect("changed", self.change_guild)
        renderer_text = Gtk.CellRendererText()
        guild.pack_start(renderer_text, True)
        guild.add_attribute(renderer_text, "text", 0)
        guild.add_attribute(renderer_text, 'sensitive', 1)

        guild_refresh = Gtk.Button.new_with_label(_("Refresh list"))
        guild_refresh.connect("pressed", self.refresh_guild_list)

        # Show Attachments
        show_attach_label = Gtk.Label.new(_("Show Attachments"))
        show_attach = Gtk.CheckButton.new()
        show_attach.set_active(self.show_attach)
        show_attach.connect("toggled", self.change_show_attach)

        # Line Limit
        line_limit_label = Gtk.Label.new(_("Line limit"))
        line_limit_label.set_xalign(0)
        line_limit_adjustment = Gtk.Adjustment.new(
            self.line_limit, 10, 400, 1, 1, 8)
        line_limit = Gtk.SpinButton.new(line_limit_adjustment, 0, 0)
        line_limit.connect("value-changed", self.change_line_limit)

        self.align_x_widget = align_x
        self.align_y_widget = align_y
        self.align_monitor_widget = monitor
        self.align_placement_widget = align_placement_button
        self.guild_widget = guild
        self.channel_widget = channel
        self.text_time_widget = text_time
        self.text_time_label_widget = text_time_label

        box.attach(enabled_label, 0, 0, 1, 1)
        box.attach(enabled, 1, 0, 1, 1)
        #box.attach(autohide_label, 0, 1, 1, 1)
        #box.attach(autohide, 1, 1, 1, 1)
        box.attach(popup_style_label, 0, 2, 1, 1)
        box.attach(popup_style, 1, 2, 1, 1)
        box.attach(text_time_label, 0, 3, 1, 1)
        box.attach(text_time, 1, 3, 1, 1)
        box.attach(guild_label, 0, 4, 1, 1)
        box.attach(guild, 1, 4, 1, 1)
        box.attach(guild_refresh, 2, 4, 1, 1)

        box.attach(channel_label, 0, 5, 1, 1)
        box.attach(channel, 1, 5, 1, 1)
        box.attach(channel_refresh, 2, 5, 1, 1)
        box.attach(font_label, 0, 6, 1, 1)
        box.attach(font, 1, 6, 1, 1)
        box.attach(fg_col_label, 0, 7, 1, 1)
        box.attach(fg_col, 1, 7, 1, 1)
        box.attach(bg_col_label, 0, 8, 1, 1)
        box.attach(bg_col, 1, 8, 1, 1)
        box.attach(align_label, 0, 9, 1, 5)
        #box.attach(align_type_box, 1, 8, 1, 1)
        box.attach(monitor, 1, 10, 1, 1)
        box.attach(align_x, 1, 11, 1, 1)
        box.attach(align_y, 1, 12, 1, 1)
        box.attach(align_placement_button, 1, 13, 1, 1)
        box.attach(show_attach_label, 0, 14, 1, 1)
        box.attach(show_attach, 1, 14, 1, 1)
        box.attach(line_limit_label, 0, 15, 1, 1)
        box.attach(line_limit, 1, 15, 1, 1)

        self.add(box)

    def change_font(self, button):
        """
        Font settings changed
        """
        font = button.get_font()
        self.overlay.set_font(font)

        self.font = font
        self.save_config()

    def change_channel(self, button):
        """
        Channel setting changed
        """
        if self.ignore_channel_change:
            return

        channel = self.channel_lookup[button.get_active()]
        self.connector.set_text_channel(channel)
        self.channel = channel
        self.save_config()

    def change_guild(self, button):
        """
        Guild setting changed
        """
        if self.ignore_guild_change:
            return
        guild_id = self.guild_lookup[button.get_active()]
        self.guild = guild_id
        self.save_config()
        self.refresh_channel_list()

    def refresh_channel_list(self, _button=None):
        self.channel_widget.set_sensitive(False)
        self.connector.request_text_rooms_for_guild(self.guild)

    def refresh_guild_list(self, _button=None):
        self.guild_widget.set_sensitive(False)
        self.channel_widget.set_sensitive(False)
        self.connector.req_guilds()
        self.connector.request_text_rooms_for_guild(self.guild)

    def change_popup_style(self, button):
        """
        Popup style setting changed
        """
        self.overlay.set_popup_style(button.get_active())

        self.popup_style = button.get_active()
        self.save_config()

        if button.get_active():
            # We're using popup style
            self.text_time_widget.show()
            self.text_time_label_widget.show()
        else:
            self.text_time_widget.hide()
            self.text_time_label_widget.hide()

    def change_text_time(self, button):
        """
        Popup style setting changed
        """
        self.overlay.set_text_time(button.get_value())

        self.text_time = button.get_value()
        self.save_config()

    def get_channel(self):
        """
        Return selected channel
        """
        return self.channel

    def set_channel(self, channel="0"):
        """
        Change the stored channel
        """
        self.channel = channel
        self.save_config()

    def get_guild(self):
        """
        Return selected guild
        """
        return self.guild

    def change_bg(self, button):
        """
        Background colour changed
        """
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.overlay.set_bg(colour)

        self.bg_col = colour
        self.save_config()

    def change_fg(self, button):
        """
        Foreground colour changed
        """
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.overlay.set_fg(colour)

        self.fg_col = colour
        self.save_config()

    def change_show_attach(self, button):
        """
        Attachment setting changed
        """
        self.overlay.set_show_attach(button.get_active())

        self.show_attach = button.get_active()
        self.save_config()

    def change_line_limit(self, button):
        """
        Line limit setting changed
        """
        self.overlay.set_line_limit(button.get_value())

        self.line_limit = button.get_value()
        self.save_config()



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
The connector for discord.
Connects in as if it was Streamkit for OBS or Xsplit and
communicates to get voice & text info to display

Terminology:
GUILDS - Often called 'Servers' in discord. It is the group of users and channels that make up
         one server.
CHANNEL - Often called 'Rooms'. Both voice and text channels are types of channel
"""
import select
import time
import json
import sys
import logging
import calendar
import websocket
import requests

log = logging.getLogger(__name__)


class DiscordConnector:
    """
    The connector for discord.
    Connects in as if it was Streamkit for OBS or Xsplit and
    communicates to get voice & text info to display
    """

    def __init__(self, discover, text_settings, voice_settings, text_overlay, voice_overlay):
        self.discover = discover
        self.text_settings = text_settings
        self.text_overlay = text_overlay
        self.voice_settings = voice_settings
        self.voice_overlay = voice_overlay
        self.websocket = None
        self.access_token = "none"
        self.oauth_token = "207646673902501888"
        self.access_delay = 0
        self.warn_connection = True
        self.error_connection = True

        self.guilds = {}
        self.channels = {}
        self.user = {}
        self.userlist = {}
        self.in_room = []
        self.current_voice = "0"
        self.current_text = "0"
        self.list_altered = False
        self.text_altered = False
        self.last_connection = ""
        self.text = []
        self.authed = False

        self.rate_limited_channels = []
        self.reconnect_delay = 0

    def get_access_token_stage1(self):
        """
        First stage of getting an access token. Request authorization from Discord client
        """
        cmd = {
            "cmd": "AUTHORIZE",
            "args":
            {
                "client_id": self.oauth_token,
                "scopes": ["rpc", "messages.read", "rpc.notifications.read"],
                "prompt": "none",
            },
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def get_access_token_stage2(self, code1):
        """
        Second stage of getting an access token. Give auth code to streamkit
        """
        url = "https://streamkit.discord.com/overlay/token"
        myobj = {"code": code1}
        response = requests.post(url, json=myobj)
        try:
            jsonresponse = json.loads(response.text)
        except json.JSONDecodeError:
            jsonresponse = {}
        if "access_token" in jsonresponse:
            self.access_token = jsonresponse["access_token"]
            self.req_auth()
        else:
            log.error("No access token in json response")
            log.error(response.text)
            self.on_close()

    def set_channel(self, channel, need_req=True):
        """
        Set currently active voice channel
        """
        if not channel:
            if self.current_voice:
                self.unsub_voice_channel(self.current_voice)
            self.current_voice = "0"
            return
        if channel != self.current_voice:
            if self.current_voice != "0":
                self.unsub_voice_channel(self.current_voice)
            self.sub_voice_channel(channel)
            self.current_voice = channel
            if need_req:
                self.req_channel_details(channel)

    def set_text_channel(self, channel, need_req=True):
        """
        Set currently active text channel
        """
        if not channel:
            self.current_text = "0"
            return
        if channel != self.current_text:
            self.start_listening_text(channel)
            if need_req:
                self.req_channel_details(channel)

    def set_in_room(self, userid, present):
        """
        Set user currently in given room
        """
        if present:
            if userid not in self.in_room:
                self.in_room.append(userid)
        else:
            if userid in self.in_room:
                self.in_room.remove(userid)

    def add_text(self, message):
        """
        Add line of text to text list. Assumes the message is from the correct room
        """
        utc_time = None
        try:
            utc_time = time.strptime(
                message["timestamp"], "%Y-%m-%dT%H:%M:%S.%f%z")
        except ValueError:
            utc_time = time.strptime(
                message["timestamp"], "%Y-%m-%dT%H:%M:%S%z")

        epoch_time = calendar.timegm(utc_time)
        username = message["author"]["username"]
        if("nick" in message and message['nick'] and len(message["nick"]) > 1
                and 'object Object' not in json.dumps(message["nick"])):
            username = message["nick"]
        colour = "#ffffff"
        if "author_color" in message:
            colour = message["author_color"]

        self.text.append({'id': message["id"],
                          'content': self.get_message_from_message(message),
                          'nick': username,
                          'nick_col': colour,
                          'time': epoch_time,
                          'attach': self.get_attachment_from_message(message),
                          })
        self.text_altered = True

    def update_text(self, message_in):
        """
        Update a line of text
        """
        for idx in range(0, len(self.text)):
            message = self.text[idx]
            if message['id'] == message_in['id']:
                new_message = {'id': message['id'],
                               'content': self.get_message_from_message(message_in),
                               'nick': message['nick'],
                               'nick_col': message['nick_col'],
                               'time': message['time'],
                               'attach': message['attach']}
                self.text[idx] = new_message
                self.text_altered = True
                return

    def delete_text(self, message_in):
        """
        Delete a line of text
        """
        for idx in range(0, len(self.text)):
            message = self.text[idx]
            if message['id'] == message_in['id']:
                del self.text[idx]
                self.text_altered = True
                return

    def get_message_from_message(self, message):
        """
        Messages are sent as JSON objects, with varying information.
        Decides which bits are shown and which are discarded
        """
        if "content_parsed" in message:
            return message["content_parsed"]
        elif "content" in message and len(message["content"]) > 0:
            return message["content"]
        elif "embeds" in message and len(message["embeds"]) == 1:
            if "rawDescription" in message["embeds"][0]:
                return message["embeds"][0]["rawDescription"]
            if "author" in message["embeds"][0]:
                return message["embeds"][0]["author"]["name"]
        elif "attachments" in message and len(message["attachments"]) == 1:
            return ""
        return ""

    def get_attachment_from_message(self, message):
        """
        Messages with attachments come in different forms, decide what is and is
        not an attachment
        """
        if "attachments" in message and len(message["attachments"]) == 1:
            return message["attachments"]
        return None

    def update_user(self, user):
        """
        Update user information
        Pass along our custom user information from version to version
        """
        if user["id"] in self.userlist:
            olduser = self.userlist[user["id"]]
            if "mute" not in user and "mute" in olduser:
                user["mute"] = olduser["mute"]
            if "deaf" not in user and "deaf" in olduser:
                user["deaf"] = olduser["deaf"]
            if "speaking" not in user and "speaking" in olduser:
                user["speaking"] = olduser["speaking"]
            if "nick" not in user and "nick" in olduser:
                user["nick"] = olduser["nick"]
            if "lastspoken" not in user and "lastspoken" in olduser:
                user["lastspoken"] = olduser["lastspoken"]
            if olduser["avatar"] != user["avatar"]:
                self.voice_overlay.delete_avatar(user["id"])
        if "lastspoken" not in user:  # Still nothing?
            user["lastspoken"] = 0  # EEEEPOOCH EEEEEPOCH! BELIEVE MEEEE
        if "speaking" not in user:
            user["speaking"] = False
        self.userlist[user["id"]] = user

    def on_message(self, message):
        """
        Recieve websocket message super-function
        """
        j = json.loads(message)
        if j["cmd"] == "AUTHORIZE":
            self.get_access_token_stage2(j["data"]["code"])
            return
        elif j["cmd"] == "DISPATCH":
            if j["evt"] == "READY":
                self.req_auth()
            elif j["evt"] == "VOICE_STATE_UPDATE":
                self.list_altered = True
                thisuser = j["data"]["user"]
                nick = j["data"]["nick"]
                thisuser["nick"] = nick
                mute = (j["data"]["voice_state"]["mute"] or
                        j["data"]["voice_state"]["self_mute"] or
                        j["data"]["voice_state"]["suppress"])
                deaf = j["data"]["voice_state"]["deaf"] or j["data"]["voice_state"]["self_deaf"]
                thisuser["mute"] = mute
                thisuser["deaf"] = deaf
                if self.current_voice != "0":
                    self.update_user(thisuser)
                self.set_in_room(thisuser["id"], True)
            elif j["evt"] == "VOICE_STATE_CREATE":
                self.list_altered = True
                thisuser = j["data"]["user"]
                nick = j["data"]["nick"]
                thisuser["nick"] = nick
                self.update_user(thisuser)
                # We've joined a room... but where?
                if j["data"]["user"]["id"] == self.user["id"]:
                    self.find_user()
            elif j["evt"] == "VOICE_STATE_DELETE":
                self.list_altered = True
                self.set_in_room(j["data"]["user"]["id"], False)
                if j["data"]["user"]["id"] == self.user["id"]:
                    self.in_room = []
            elif j["evt"] == "SPEAKING_START":
                self.list_altered = True
                # It's only possible to get alerts for the room you're in
                self.set_channel(j["data"]["channel_id"])
                self.userlist[j["data"]["user_id"]]["speaking"] = True
                self.userlist[j["data"]["user_id"]]["lastspoken"] = time.time()
                self.list_altered = True
                self.set_in_room(j["data"]["user_id"], True)
            elif j["evt"] == "SPEAKING_STOP":
                self.list_altered = True
                # It's only possible to get alerts for the room you're in
                self.set_channel(j["data"]["channel_id"])
                if j["data"]["user_id"] in self.userlist:
                    self.userlist[j["data"]["user_id"]]["speaking"] = False
                self.set_in_room(j["data"]["user_id"], True)
            elif j["evt"] == "VOICE_CHANNEL_SELECT":
                self.set_channel(j["data"]["channel_id"])
            elif j["evt"] == "VOICE_CONNECTION_STATUS":
                # VOICE_CONNECTED > CONNECTING > AWAITING_ENDPOINT > DISCONNECTED
                self.last_connection = j["data"]["state"]
            elif j["evt"] == "MESSAGE_CREATE":
                if self.current_text == j["data"]["channel_id"]:
                    self.add_text(j["data"]["message"])
            elif j["evt"] == "MESSAGE_UPDATE":
                if self.current_text == j["data"]["channel_id"]:
                    self.update_text(j["data"]["message"])
            elif j["evt"] == "MESSAGE_DELETE":
                if self.current_text == j["data"]["channel_id"]:
                    self.delete_text(j["data"]["message"])
            elif j["evt"] == "CHANNEL_CREATE":
                # We haven't been told what guild this is in
                self.req_channel_details(j["data"]["id"], 'new')
            elif j["evt"] == "NOTIFICATION_CREATE":
                self.discover.notification_overlay.add_notification_message(j)
            else:
                log.warning(j)
            return
        elif j["cmd"] == "AUTHENTICATE":
            if j["evt"] == "ERROR":
                self.get_access_token_stage1()
                return
            else:
                self.req_guilds()
                self.user = j["data"]["user"]
                log.info(
                    "ID is %s", self.user["id"])
                log.info(
                    "Logged in as %s", self.user["username"])
                self.authed = True
                self.on_connected()
                return
        elif j["cmd"] == "GET_GUILDS":
            for guild in j["data"]["guilds"]:
                self.guilds[guild["id"]] = guild
                if len(self.voice_settings.guild_ids) == 0 or guild["id"] in self.voice_settings.guild_ids:
                    self.req_channels(guild["id"])
            return
        elif j["cmd"] == "GET_GUILD":
            # We currently only get here because of a "CHANNEL_CREATE" event. Stupidly long winded way around
            if j["data"]:
                guild = j["data"]
                if len(self.voice_settings.guild_ids) == 0 or guild["id"] in self.voice_settings.guild_ids:
                    self.req_channels(guild["id"])
            return
        elif j["cmd"] == "GET_CHANNELS":
            self.guilds[j['nonce']]["channels"] = j["data"]["channels"]
            for channel in j["data"]["channels"]:
                channel['guild_id'] = j['nonce']
                channel['guild_name'] = self.guilds[j['nonce']]["name"]
                self.channels[channel["id"]] = channel
                if channel["type"] == 2:
                    self.req_channel_details(channel["id"])
            if j["nonce"] == self.text_settings.get_guild():
                self.text_settings.set_channels(j["data"]["channels"])
            self.check_guilds()
            return
        elif j["cmd"] == "SUBSCRIBE":
            # Only log errors
            if j['evt']:
                log.warning(j)
            return
        elif j["cmd"] == "UNSUBSCRIBE":
            return
        elif j["cmd"] == "GET_SELECTED_VOICE_CHANNEL":
            if 'data' in j and j['data'] and 'id' in j['data']:
                self.set_channel(j['data']['id'])
                self.list_altered = True
                self.in_room = []
                for u in j['data']['voice_states']:
                    thisuser = u["user"]
                    nick = u["nick"]
                    thisuser["nick"] = nick
                    mute = (u["voice_state"]["mute"] or
                            u["voice_state"]["self_mute"] or
                            u["voice_state"]["suppress"])
                    deaf = u["voice_state"]["deaf"] or u["voice_state"]["self_deaf"]
                    thisuser["mute"] = mute
                    thisuser["deaf"] = deaf
                    self.update_user(thisuser)
                    self.set_in_room(thisuser["id"], True)
            return
        elif j["cmd"] == "GET_CHANNEL":
            if j["evt"] == "ERROR":
                log.info(
                    "Could not get room")
                return
            if j["nonce"] == "new":
                self.req_channels(j["data"]["guild_id"])
            if j["data"]["type"] == 0:  # Text channel
                if self.current_text == j["data"]["id"]:
                    self.text = []
                    for message in j["data"]["messages"]:
                        self.add_text(message)

            return
        elif j["cmd"] == "SELECT_VOICE_CHANNEL":
            return
        elif j["cmd"] == "SET_VOICE_SETTINGS":
            return
        elif j["cmd"] == "GET_VOICE_SETTINGS":
            return
        log.warning(j)

    def check_guilds(self):
        """
        Check if all of the guilds contain a channel
        """
        for guild in self.guilds.values():
            if len(self.voice_settings.guild_ids) > 0 and guild["id"] in self.voice_settings.guild_ids and "channels" not in guild:
                return

    def on_connected(self):
        """
        Called when connection is finalised
        """
        self.sub_server()
        self.find_user()
        self.voice_overlay.set_enabled(True)
        if self.text_overlay:
            self.text_overlay.set_enabled(self.text_settings.enabled)
            if self.current_text:
                self.start_listening_text(self.current_text)

    def on_error(self, error):
        """
        Called when an error has occured
        """
        log.error("ERROR : %s", error)

    def on_close(self):
        """
        Called when connection is closed
        """
        log.warning("Connection closed")
        self.voice_overlay.hide()
        if self.text_overlay:
            self.text_overlay.hide()
        self.websocket = None
        self.reconnect_delay = 60 * 5

    def req_auth(self):
        """
        Request authentication token
        """
        cmd = {
            "cmd": "AUTHENTICATE",
            "args": {
                "access_token": self.access_token
            },
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def req_guild(self, guild_id, nonce):
        """
        Request info on one guild
        """
        cmd = {
            "cmd": "GET_GUILD",
            "args": {"guild_id": guild_id},
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def req_guilds(self):
        """
        Request all guilds information for logged in user
        """
        cmd = {
            "cmd": "GET_GUILDS",
            "args": {},
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def req_channels(self, guild):
        """
        Request all channels information for given guild.
        Don't perform now but pass off to rate-limiter
        """

        if guild in self.guilds:
            self.rate_limited_channels.append(guild)
        else:
            log.warning(f"Didn't find guild with id {guild}")

    def req_channel_details(self, channel, nonce=None):
        """message
        Request information about a specific channel
        """
        if not self.websocket:
            return
        if not nonce:
            nonce = channel
        cmd = {
            "cmd": "GET_CHANNEL",
            "args": {
                "channel_id": channel
            },
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def find_user(self):
        """
        Find the user
        """

        cmd = {
            "cmd": "GET_SELECTED_VOICE_CHANNEL",
            "args": {

            },
            "nonce": "test"
        }
        self.websocket.send(json.dumps(cmd))

    def sub_raw(self, event, args, nonce):
        """
        Subscribe to event helper function
        """
        cmd = {
            "cmd": "SUBSCRIBE",
            "args": args,
            "evt": event,
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def unsub_raw(self, event, args, nonce):
        """
        Subscribe to event helper function
        """
        cmd = {
            "cmd": "UNSUBSCRIBE",
            "args": args,
            "evt": event,
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def sub_server(self):
        """
        Subscribe to helpful events that report connectivity issues &
        when the user has intentionally changed channel

        Unfortunatly no event has been found to alert to being forcibly moved
        or that reports the users current location
        """
        self.sub_raw("VOICE_CHANNEL_SELECT", {}, "VOICE_CHANNEL_SELECT")
        self.sub_raw("VOICE_CONNECTION_STATUS", {}, "VOICE_CONNECTION_STATUS")
        self.sub_raw("GUILD_CREATE", {}, "GUILD_CREATE")
        self.sub_raw("CHANNEL_CREATE", {}, "CHANNEL_CREATE")
        self.sub_raw("NOTIFICATION_CREATE", {}, "NOTIFICATION_CREATE")

    def sub_channel(self, event, channel):
        """
        Subscribe to event on channel
        """
        self.sub_raw(event, {"channel_id": channel}, channel)

    def unsub_channel(self, event, channel):
        """
        Subscribe to event on channel
        """
        self.unsub_raw(event, {"channel_id": channel}, channel)

    def sub_text_channel(self, channel):
        """
        Subscribe to text-based events.
        """
        self.sub_channel("MESSAGE_CREATE", channel)
        self.sub_channel("MESSAGE_UPDATE", channel)
        self.sub_channel("MESSAGE_DELETE", channel)

    def unsub_text_channel(self, channel):
        """
        Unsubscribe to text-based events.
        """
        self.unsub_channel("MESSAGE_CREATE", channel)
        self.unsub_channel("MESSAGE_UPDATE", channel)
        self.unsub_channel("MESSAGE_DELETE", channel)

    def sub_voice_channel(self, channel):
        """
        Subscribe to voice-based events
        """
        self.sub_channel("VOICE_STATE_CREATE", channel)
        self.sub_channel("VOICE_STATE_UPDATE", channel)
        self.sub_channel("VOICE_STATE_DELETE", channel)
        self.sub_channel("SPEAKING_START", channel)
        self.sub_channel("SPEAKING_STOP", channel)

    def unsub_voice_channel(self, channel):
        """
        Remove subscription to voice-based events
        """
        self.unsub_channel("VOICE_STATE_CREATE", channel)
        self.unsub_channel("VOICE_STATE_UPDATE", channel)
        self.unsub_channel("VOICE_STATE_DELETE", channel)
        self.unsub_channel("SPEAKING_START", channel)
        self.unsub_channel("SPEAKING_STOP", channel)

    def get_voice_settings(self):
        """
        Request a recent version of voice settings
        """
        cmd = {
            "cmd": "GET_VOICE_SETTINGS",
            "args": {},
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def set_mute(self, muted):
        cmd = {
            "cmd": "SET_VOICE_SETTINGS",
            "args": {"mute": muted},
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def set_deaf(self, deaf):
        cmd = {
            "cmd": "SET_VOICE_SETTINGS",
            "args": {"deaf": deaf},
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def change_voice_room(self, id):
        """
        Switch to another voice room
        """
        cmd = {
            "cmd": "SELECT_VOICE_CHANNEL",
            "args": {
                "channel_id": id,
                "force": True
            },
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def change_text_room(self, id):
        """
        Switch to another text room
        """
        cmd = {
            "cmd": "SELECT_TEXT_CHANNEL",
            "args": {
                "channel_id": id
            },
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def do_read(self):
        """
        Poorly named logic center.

        Checks for new data on socket, passes to on_message

        Also passes out text data to text overlay and voice data to voice overlay

        Called at 60Hz approximately but has near zero bearing on rendering
        """
        if self.discover.show_settings_delay:
            self.discover.show_settings_delay = False
            self.discover.show_settings()
        # Ensure connection
        if not self.websocket:
            if self.reconnect_delay <= 0:
                # No timeout left, connect to discord again
                self.connect()
                if self.warn_connection:
                    log.warning(
                        "Unable to connect to Discord client")
                    self.warn_connection = False
                return True
            else:
                # Timeout requested, wait it out
                self.reconnect_delay -= 1
                return True
        # Recreate a list of users in current room
        newlist = []
        for userid in self.in_room:
            newlist.append(self.userlist[userid])
        self.voice_overlay.set_user_list(newlist, self.list_altered)
        self.voice_overlay.set_connection(self.last_connection)
        self.list_altered = False
        # Update text list
        if self.text_overlay.popup_style:
            self.text_altered = True
        if self.text_altered:
            self.text_overlay.set_text_list(self.text, self.text_altered)
            self.text_altered = False
        # Update guilds
        self.text_settings.set_guilds(self.guilds)
        # Check for changed channel
        if self.authed:
            self.set_text_channel(self.text_settings.get_channel())

        if len(self.rate_limited_channels) > 0:
            guild = self.rate_limited_channels.pop()
            cmd = {
                "cmd": "GET_CHANNELS",
                "args": {
                    "guild_id": guild
                },
                "nonce": guild
            }
            self.websocket.send(json.dumps(cmd))

        # Poll socket for new information
        recv, _w, _e = select.select((self.websocket.sock,), (), (), 0)
        while recv:
            try:
                # Receive & send to on_message
                msg = self.websocket.recv()
                self.on_message(msg)
                if not self.websocket:
                    # Connection was closed in the meantime
                    return True
                recv, _w, _e = select.select((self.websocket.sock,), (), (), 0)
            except websocket.WebSocketConnectionClosedException:
                self.on_close()
                return True
        return True

    def start_listening_text(self, channel):
        """
        Subscribe to text events on channel, or remember the channel for when we've connected

        Helper function to avoid race conditions of reading config vs connecting to websocket
        """
        if self.websocket:
            if self.current_text != "0":
                self.unsub_text_channel(self.current_text)
            if channel != "0":
                self.sub_text_channel(channel)
                self.req_channel_details(channel)
        self.current_text = channel

    def request_text_rooms_for_guild(self, guild_id):
        """
        Request a correctly ordered list of text channels.

        This will be mixed in with 'None' in the list where a voice channel is
        """
        if(guild_id == 0):
            return
        if guild_id in self.guilds:
            self.req_guild(guild_id, "refresh")

    def connect(self):
        """
        Attempt to connect to websocket

        Should not throw simply for being unable to connect, only for more serious issues
        """
        if self.websocket:
            return
        try:
            self.websocket = websocket.create_connection(
                "ws://127.0.0.1:6463/?v=1&client_id=%s" % (self.oauth_token),
                origin="https://streamkit.discord.com"
            )
        except ConnectionError as error:
            if self.error_connection:
                log.error(error)
                self.error_connection = False



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""An X11 window which can be moved and resized"""
import gi
import cairo
import logging
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position
from gi.repository import Gtk, Gdk  # nopep8

log = logging.getLogger(__name__)

class DraggableWindow(Gtk.Window):
    """An X11 window which can be moved and resized"""

    def __init__(self, pos_x=0, pos_y=0, width=300, height=300, message="Message", settings=None):
        Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.width = max(100, width)
        self.height = max(100, height)
        self.settings = settings
        self.message = message
        self.set_size_request(50, 50)

        self.connect('draw', self.dodraw)
        self.connect('motion-notify-event', self.drag)
        self.connect('button-press-event', self.button_press)
        self.connect('button-release-event', self.button_release)

        self.compositing = False
        # Set RGBA
        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        if visual:
            # Set the visual even if we can't use it right now
            self.set_visual(visual)
        if screen.is_composited():
            self.compositing = True

        self.set_app_paintable(True)
        self.monitor = 0

        self.drag_type = None
        self.drag_x = 0
        self.drag_y = 0
        self.force_location()
        self.show_all()
    
    def force_location(self):
        """
        Move the window to previously given co-ords.
        Also double check sanity on layer & decorations
        """
        self.set_decorated(False)
        self.set_keep_above(True)
        self.move(self.pos_x, self.pos_y)
        self.resize(self.width, self.height)

    def drag(self, _w, event):
        """Called by GTK while mouse is moving over window. Used to resize and move"""
        if event.state & Gdk.ModifierType.BUTTON1_MASK:
            if self.drag_type == 1:
                # Center is move
                self.pos_x = event.x_root - self.drag_x
                self.pos_y = event.y_root - self.drag_y
                self.force_location()
            elif self.drag_type == 2:
                # Right edge
                self.width += event.x - self.drag_x
                self.drag_x = event.x
                self.force_location()
            elif self.drag_type == 3:
                # Bottom edge
                self.height += event.y - self.drag_y
                self.drag_y = event.y
                self.force_location()
            else:
                # Bottom Right
                self.width += event.x - self.drag_x
                self.height += event.y - self.drag_y
                self.drag_x = event.x
                self.drag_y = event.y
                self.force_location()

    def button_press(self, _widget, event):
        """Called when a mouse button is pressed on this window"""
        (width, height) = self.get_size()
        if not self.drag_type:
            self.drag_type = 1
            # Where in the window did we press?
            if event.y > height - 32:
                self.drag_type += 2
            if event.x > width - 32:
                self.drag_type += 1
            self.drag_x = event.x
            self.drag_y = event.y

    def button_release(self, _w, _event):
        """Called when a mouse button is released"""
        self.drag_type = None

    def dodraw(self, _widget, context):
        """Draw our window."""
        context.set_source_rgba(1.0, 1.0, 0.0, 0.7)
        # Don't layer drawing over each other, always replace
        context.set_operator(cairo.OPERATOR_SOURCE)
        context.paint()
        context.set_operator(cairo.OPERATOR_OVER)
        # Get size of window
        (window_width, window_height) = self.get_size()

        # Draw text
        context.set_source_rgba(0.0, 0.0, 0.0, 1.0)
        _xb, _yb, text_width, text_height, _dx, _dy = context.text_extents(
            self.message)
        context.move_to(window_width / 2 - text_width / 2,
                        window_height / 2 - text_height / 2)
        context.show_text(self.message)

        # Draw resizing edges
        context.set_source_rgba(0.0, 0.0, 1.0, 0.5)
        context.rectangle(window_width - 32, 0, 32, window_height)
        context.fill()

        context.rectangle(0, window_height - 32, window_width, 32)
        context.fill()

    def get_coords(self):
        """Return window position and size"""
        scale = self.get_scale_factor()
        (pos_x, pos_y) = self.get_position()
        (width, height) = self.get_size()
        pos_x = pos_x / scale
        pos_y = pos_y / scale
        return (pos_x, pos_y, width, height)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Overlay window for text"""
import logging
import time
import re
import cairo
import gi
from .image_getter import get_surface, draw_img_to_rect, get_aspected_size
from .overlay import OverlayWindow
gi.require_version("Gtk", "3.0")
gi.require_version('PangoCairo', '1.0')
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Pango, PangoCairo  # nopep8

log = logging.getLogger(__name__)


class TextOverlayWindow(OverlayWindow):
    """Overlay window for text"""

    def __init__(self, discover, piggyback=None):
        OverlayWindow.__init__(self, discover, piggyback)
        self.text_spacing = 4
        self.content = []
        self.text_font = None
        self.text_size = 13
        self.text_time = None
        self.show_attach = None
        self.popup_style = None
        self.line_limit = 100
        # 0, 0, self.text_size, self.text_size)
        self.pango_rect = Pango.Rectangle()
        self.pango_rect.width = self.text_size * Pango.SCALE
        self.pango_rect.height = self.text_size * Pango.SCALE

        self.connected = True
        self.bg_col = [0.0, 0.6, 0.0, 0.1]
        self.fg_col = [1.0, 1.0, 1.0, 1.0]
        self.attachment = {}

        self.image_list = []
        self.img_finder = re.compile(r"`")
        self.warned_filetypes = []
        self.set_title("Discover Text")
        self.redraw()

    def tick(self):
        if len(self.attachment) > self.line_limit:
            # We've probably got old images!
            oldlist = self.attachment
            self.attachment = {}
            log.info("Cleaning old images")
            for message in self.content:
                if 'attach' in message and message['attach']:
                    url = message['attach'][0]['url']
                    log.info("keeping %s", url)
                    self.attachment[url] = oldlist[url]

    def set_text_time(self, timer):
        """
        Set the duration that a message will be visible for.
        """
        self.text_time = timer
        self.needsredraw = True

    def set_text_list(self, tlist, altered):
        """
        Update the list of text messages to show
        """
        self.content = tlist[-self.line_limit:]
        if altered:
            self.needsredraw = True

    def set_fg(self, fg_col):
        """
        Set default text colour
        """
        self.fg_col = fg_col
        self.needsredraw = True

    def set_bg(self, bg_col):
        """
        Set background colour
        """
        self.bg_col = bg_col
        self.needsredraw = True

    def set_show_attach(self, attachment):
        """
        Set if attachments should be shown inline
        """
        self.show_attach = attachment
        self.needsredraw = True

    def set_popup_style(self, boolean):
        """
        Set if message disappear after a certain duration
        """
        self.popup_style = boolean

    def set_font(self, font):
        """
        Set font used to render text
        """
        self.text_font = font

        self.pango_rect = Pango.Rectangle()
        font = Pango.FontDescription(self.text_font)
        self.pango_rect.width = font.get_size() * Pango.SCALE
        self.pango_rect.height = font.get_size() * Pango.SCALE
        self.needsredraw = True

    def set_line_limit(self, limit):
        """
        Change maximum number of lines in overlay
        """
        self.line_limit = limit

    def make_line(self, message):
        """
        Decode a recursive JSON object into pango markup.
        """
        ret = ""
        if isinstance(message, list):
            for inner_message in message:
                ret = "%s%s" % (ret, self.make_line(inner_message))
        elif isinstance(message, str):
            ret = self.sanitize_string(message)
        elif message['type'] == 'strong':
            ret = "<b>%s</b>" % (self.make_line(message['content']))
        elif message['type'] == 'text':
            ret = self.sanitize_string(message['content'])
        elif message['type'] == 'link':
            ret = "<u>%s</u>" % (self.make_line(message['content']))
        elif message['type'] == 'emoji':
            if 'surrogate' in message:
                # ['src'] is SVG URL
                # ret = msg
                ret = message['surrogate']
            else:
                ### Add Image ###
                self.image_list.append(
                    f"https://cdn.discordapp.com/emojis/{message['emojiId']}.png?v=1"
                )
                ret = "`"
        elif (message['type'] == 'inlineCode' or
              message['type'] == 'codeBlock' or
              message['type'] == 'blockQuote'):
            ret = "<span font_family=\"monospace\" background=\"#0004\">%s</span>" % (
                self.make_line(message['content']))
        elif message['type'] == 'u':
            ret = "<u>%s</u>" % (self.make_line(message['content']))
        elif message['type'] == 'em':
            ret = "<i>%s</i>" % (self.make_line(message['content']))
        elif message['type'] == 's':
            ret = "<s>%s</s>" % (self.make_line(message['content']))
        elif message['type'] == 'channel':
            ret = self.make_line(message['content'])
        elif message['type'] == 'mention':
            ret = self.make_line(message['content'])
        elif message['type'] == 'br':
            ret = '\n'
        else:
            if message["type"] not in self.warned_filetypes:
                log.error("Unknown text type : %s", message["type"])
                self.warned_filetypes.append(message['type'])
        return ret

    def recv_attach(self, identifier, pix):
        """
        Called when an image has been downloaded by image_getter
        """
        self.attachment[identifier] = pix
        self.needsredraw = True

    def overlay_draw(self, w, context, data=None):
        """
        Draw the overlay
        """
        if self.piggyback:
            self.piggyback.overlay_draw(w, context, data)
        if not self.enabled:
            return
        self.context = context
        (width, height) = self.get_size()
        if not self.piggyback_parent:
            context.set_antialias(cairo.ANTIALIAS_GOOD)
            context.set_source_rgba(0.0, 0.0, 0.0, 0.0)
            context.set_operator(cairo.OPERATOR_SOURCE)
            context.paint()
        context.save()
        if self.is_wayland or self.piggyback_parent or self.discover.steamos:
            # Special case!
            # The window is full-screen regardless of what the user has selected.
            # We need to set a clip and a transform to imitate original behaviour
            # Used in wlroots & gamescope
            width = self.width
            height = self.height
            context.translate(self.pos_x, self.pos_y)
            context.rectangle(0, 0, width, height)
            context.clip()

        current_y = height
        tnow = time.time()
        for line in reversed(self.content):
            if self.popup_style and tnow - line['time'] > self.text_time:
                break
            out_line = ""
            self.image_list = []

            col = "#fff"
            if 'nick_col' in line and line['nick_col']:
                col = line['nick_col']
            for in_line in line['content']:
                out_line = "%s%s" % (out_line, self.make_line(in_line))
            if line['attach'] and self.show_attach:
                attachment = line['attach'][0]
                url = attachment['url']
                extension = attachment['filename']
                extension = extension.rsplit(".", 1)[1]
                extension = extension.lower()
                if extension in ['jpeg', 'jpg', 'png', 'gif']:
                    if url in self.attachment:
                        current_y = self.draw_attach(current_y, url)
                    else:
                        get_surface(self.recv_attach,
                                    url,
                                    url, None)
                        self.attachment[url] = None  # Avoid asking repeatedly
                else:
                    log.warning("Unknown file extension '%s'", extension)
                # cy = self.draw_text(cy, "%s" % (line['attach']))
            message = "<span foreground='%s'>%s</span>: %s" % (self.sanitize_string(col),
                                                               self.sanitize_string(
                                                                   line["nick"]),
                                                               out_line)
            current_y = self.draw_text(current_y, message)
            if current_y <= 0:
                # We've done enough
                break
        context.restore()
        self.context = None

    def draw_attach(self, pos_y, url):
        """
        Draw an attachment
        """
        if url in self.attachment and self.attachment[url]:
            pix = self.attachment[url]
            image_width = min(pix.get_width(), self.width)
            image_height = min(pix.get_height(), (self.height * .7))
            (_ax, _ay, _aw, aspect_height) = get_aspected_size(
                pix, image_width, image_height)
            self.col(self.bg_col)
            self.context.rectangle(0, pos_y - aspect_height,
                                   self.width, aspect_height)

            self.context.fill()
            self.context.set_operator(cairo.OPERATOR_OVER)
            _new_w, new_h = draw_img_to_rect(
                pix, self.context, 0, pos_y - image_height, image_width, image_height, aspect=True)
            return pos_y - new_h
        return pos_y

    def draw_text(self, pos_y, text):
        """
        Draw a text message, returning the Y position of the next message
        """
        layout = self.create_pango_layout(text)
        layout.set_auto_dir(True)
        layout.set_markup(text, -1)
        attr = layout.get_attributes()

        layout.set_width(Pango.SCALE * self.width)
        layout.set_spacing(Pango.SCALE * 3)
        if self.text_font:
            font = Pango.FontDescription(self.text_font)
            layout.set_font_description(font)
        _tw, text_height = layout.get_pixel_size()
        self.col(self.bg_col)
        self.context.rectangle(0, pos_y - text_height, self.width, text_height)
        self.context.fill()
        self.context.set_operator(cairo.OPERATOR_OVER)
        self.col(self.fg_col)

        self.context.move_to(0, pos_y - text_height)
        PangoCairo.context_set_shape_renderer(
            self.get_pango_context(), self.render_custom, None)

        text = layout.get_text()
        count = 0

        for loc in self.img_finder.finditer(text):
            idx = loc.start()

            if len(self.image_list) <= count:
                break  # We fucked up. Who types ` anyway
            # url = self.imgList[count]

            attachment = Pango.attr_shape_new_with_data(
                self.pango_rect, self.pango_rect, count, None)
            attachment.start_index = idx
            attachment.end_index = idx + 1
            attr.insert(attachment)
            count += 1
        layout.set_attributes(attr)

        PangoCairo.show_layout(self.context, layout)
        return pos_y - text_height

    def render_custom(self, ctx, shape, path, _data):
        """
        Draw an inline image as a custom emoticon
        """
        if shape.data >= len(self.image_list):
            log.warning(f"{shape.data} >= {len(self.image_list)}")
            return
        # key is the url to the image
        key = self.image_list[shape.data]
        if key not in self.attachment:
            get_surface(self.recv_attach,
                        key,
                        key, None)
            return
        pix = self.attachment[key]
        (pos_x, pos_y) = ctx.get_current_point()
        draw_img_to_rect(pix, ctx, pos_x, pos_y - self.text_size, self.text_size,
                         self.text_size, path=path)
        return True

    def sanitize_string(self, string):
        """
        Sanitize a text message so that it doesn't intefere with Pango's XML format
        """
        string = string.replace("&", "&amp;")
        string = string.replace("<", "&lt;")
        string = string .replace(">", "&gt;")
        string = string.replace("'", "&#39;")
        string = string.replace("\"", "&#34;")
        return string



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
Settings tab parent class. Helpful if we need more
overlay types without copy-and-pasting too much code
"""
import gettext
import os
import logging
import pkg_resources
import gi
from .draggable_window import DraggableWindow
from .draggable_window_wayland import DraggableWindowWayland
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Gdk  # nopep8


try:
    from xdg.BaseDirectory import xdg_config_home
except ModuleNotFoundError:
    from xdg import XDG_CONFIG_HOME as xdg_config_home

log = logging.getLogger(__name__)
t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'))
_ = t.gettext


class SettingsWindow(Gtk.VBox):
    """
    Settings tab parent class. Helpful if we need more
    overlay types without copy-and-pasting too much code
    """

    def __init__(self, discover):
        Gtk.VBox.__init__(self)
        self.placement_window = None
        self.config_dir = None
        self.config_file = None
        self.overlay = None
        self.floating = None
        self.floating_x = None
        self.floating_y = None
        self.floating_w = None
        self.floating_h = None
        self.align_x_widget = None
        self.align_y_widget = None
        self.align_monitor_widget = None
        self.align_placement_widget = None
        self.monitor = None
        self.align_x = None
        self.align_y = None
        self.enabled = None
        self.autohide = None
        self.discover = discover

    def init_config(self):
        """
        Locate the config and then read
        """
        self.config_dir = os.path.join(xdg_config_home, "discover_overlay")
        os.makedirs(self.config_dir, exist_ok=True)
        self.config_file = os.path.join(self.config_dir, "config.ini")
        self.read_config()

    def close_window(self, _a=None, _b=None):
        """
        Helper to ensure we don't lose changes to floating windows
        Hide for later
        """
        if self.placement_window:
            (pos_x, pos_y, width, height) = self.placement_window.get_coords()

            self.floating_x = pos_x
            self.floating_y = pos_y
            self.floating_w = width
            self.floating_h = height
            self.overlay.set_floating(True, pos_x, pos_y, width, height)
            self.save_config()
            self.placement_window.close()
            self.placement_window = None
        self.hide()
        return True

    def get_monitor_index(self, name):
        """
        Helper function to find the index number of the monitor
        """
        display = Gdk.Display.get_default()
        if "get_n_monitors" in dir(display):
            for i in range(0, display.get_n_monitors()):
                if display.get_monitor(i).get_model() == name:
                    return i
        return 0

    def get_monitor_obj(self, name):
        """
        Helper function to find the monitor object of the monitor
        """
        display = Gdk.Display.get_default()
        if "get_n_monitors" in dir(display):
            for i in range(0, display.get_n_monitors()):
                if display.get_monitor(i).get_model() == name:
                    return display.get_monitor(i)

        return None

    def present_settings(self):
        """
        Show settings
        """
        self.show_all()

    def read_config(self):
        """
        Stub called when settings are needed to be read
        """

    def save_config(self):
        """
        Stub called when settings are needed to be written
        """

    def change_placement(self, button):
        """
        Placement window button pressed.
        """
        if self.placement_window:
            (pos_x, pos_y, width, height) = self.placement_window.get_coords()
            self.floating_x = pos_x
            self.floating_y = pos_y
            self.floating_w = width
            self.floating_h = height

            log.info("Positioned overlay : %s , %s  %s x %s", self.floating_x,
                     self.floating_y, self.floating_w, self.floating_h)
            self.overlay.set_floating(True, pos_x, pos_y, width, height)
            self.save_config()
            if button:
                button.set_label(_("Place Window"))
            self.placement_window.close()
            self.placement_window = None
            if self.discover.steamos:
                self.discover.show_settings()
        else:
            if self.discover.steamos:
                self.discover.hide_settings()
            if self.overlay.is_wayland or self.discover.steamos:
                self.placement_window = DraggableWindowWayland(
                    pos_x=self.floating_x, pos_y=self.floating_y,
                    width=self.floating_w, height=self.floating_h,
                    message=_("Place & resize this window then press Green!"), settings=self,
                    steamos=self.discover.steamos)
            else:
                self.placement_window = DraggableWindow(
                    pos_x=self.floating_x, pos_y=self.floating_y,
                    width=self.floating_w, height=self.floating_h,
                    message=_("Place & resize this window then press Save!"), settings=self)
                if button:
                    button.set_label(_("Save this position"))

    def change_align_type_edge(self, button):
        """
        Alignment setting changed
        """
        if button.get_active():
            self.overlay.set_floating(
                False, self.floating_x, self.floating_y, self.floating_w, self.floating_h)
            self.floating = False
            self.save_config()

            # Re-sort the screen
            self.align_x_widget.show()
            self.align_y_widget.show()
            self.align_monitor_widget.show()
            self.align_placement_widget.hide()

    def change_align_type_floating(self, button):
        """
        Alignment setting changed
        """
        if button.get_active():
            self.overlay.set_floating(
                True, self.floating_x, self.floating_y, self.floating_w, self.floating_h)
            self.floating = True
            self.save_config()
            self.align_x_widget.hide()
            self.align_y_widget.hide()
            self.align_monitor_widget.hide()
            self.align_placement_widget.show()

    def change_monitor(self, button):
        """
        Alignment setting changed
        """
        display = Gdk.Display.get_default()
        if "get_monitor" in dir(display):
            mon = display.get_monitor(button.get_active())
            m_s = mon.get_model()
            self.overlay.set_monitor(button.get_active(), mon)

            self.monitor = m_s
            self.save_config()

    def change_align_x(self, button):
        """
        Alignment setting changed
        """
        self.overlay.set_align_x(button.get_active() == 1)

        self.align_x = (button.get_active() == 1)
        self.save_config()

    def change_align_y(self, button):
        """
        Alignment setting changed
        """
        self.overlay.set_align_y(button.get_active())

        self.align_y = button.get_active()
        self.save_config()

    def change_enabled(self, button):
        """
        Overlay active state toggled
        """
        self.overlay.set_enabled(button.get_active())

        self.enabled = button.get_active()
        self.save_config()

    def change_hide_on_mouseover(self, button):
        """
        Autohide setting changed
        """
        self.overlay.set_hide_on_mouseover(button.get_active())
        self.autohide = button.get_active()
        self.save_config()



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Notification setting tab on settings window"""
import gettext
import json
import pkg_resources
from configparser import ConfigParser
import gi
from .settings import SettingsWindow

gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Gdk  # nopep8


GUILD_DEFAULT_VALUE = "0"
t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'))
_ = t.gettext


class NotificationSettingsWindow(SettingsWindow):
    """Notification setting tab on settings window"""

    def __init__(self, overlay, discover):
        SettingsWindow.__init__(self, discover)
        self.overlay = overlay

        self.placement_window = None
        self.init_config()
        self.align_x = None
        self.align_y = None
        self.monitor = None
        self.floating = None
        self.font = None
        self.bg_col = None
        self.fg_col = None
        self.text_time = None
        self.show_icon = None
        self.enabled = None
        self.limit_width = None
        self.icon_left = None
        self.padding = None
        self.icon_padding = None
        self.icon_size = None
        self.border_radius = None

        self.set_size_request(400, 200)
        self.connect("destroy", self.close_window)
        self.connect("delete-event", self.close_window)
        if overlay:
            self.init_config()
            self.create_gui()

    def present_settings(self):
        """
        Show contents of tab and update lists
        """
        if not self.overlay:
            return
        self.show_all()
        if not self.floating:
            self.align_x_widget.show()
            self.align_y_widget.show()
            self.align_monitor_widget.show()
            self.align_placement_widget.hide()
        else:
            self.align_x_widget.hide()
            self.align_y_widget.hide()
            self.align_monitor_widget.show()
            self.align_placement_widget.show()

    def read_config(self):
        """
        Read in the 'text' section of the config
        """
        if not self.overlay:
            return
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        self.enabled = config.getboolean(
            "notification", "enabled", fallback=False)
        self.align_x = config.getboolean(
            "notification", "rightalign", fallback=True)
        self.align_y = config.getint("notification", "topalign", fallback=2)
        self.monitor = config.get("notification", "monitor", fallback="None")
        self.floating = config.getboolean(
            "notification", "floating", fallback=False)
        self.floating_x = config.getint(
            "notification", "floating_x", fallback=0)
        self.floating_y = config.getint(
            "notification", "floating_y", fallback=0)
        self.floating_w = config.getint(
            "notification", "floating_w", fallback=400)
        self.floating_h = config.getint(
            "notification", "floating_h", fallback=400)
        self.font = config.get("notification", "font", fallback=None)
        self.bg_col = json.loads(config.get(
            "notification", "bg_col", fallback="[0.0,0.0,0.0,0.5]"))
        self.fg_col = json.loads(config.get(
            "notification", "fg_col", fallback="[1.0,1.0,1.0,1.0]"))
        self.text_time = config.getint(
            "notification", "text_time", fallback=10)
        self.show_icon = config.getboolean(
            "notification", "show_icon", fallback=True)
        self.reverse_order = config.getboolean(
            "notification", "rev", fallback=False)
        self.limit_width = config.getint(
            "notification", "limit_width", fallback=400)
        self.icon_left = config.getboolean(
            "notification", "icon_left", fallback=True)
        self.icon_padding = config.getint(
            "notification", "icon_padding", fallback=8)
        self.icon_size = config.getint(
            "notification", "icon_size", fallback=32)
        self.padding = config.getint(
            "notification", "padding", fallback=8)
        self.border_radius = config.getint(
            "notification", "border_radius", fallback=8)

        # Pass all of our config over to the overlay
        self.overlay.set_enabled(self.enabled)
        self.overlay.set_align_x(self.align_x)
        self.overlay.set_align_y(self.align_y)
        self.overlay.set_monitor(self.get_monitor_index(
            self.monitor), self.get_monitor_obj(self.monitor))
        self.overlay.set_floating(
            self.floating, self.floating_x, self.floating_y, self.floating_w, self.floating_h)
        self.overlay.set_bg(self.bg_col)
        self.overlay.set_fg(self.fg_col)
        self.overlay.set_text_time(self.text_time)
        self.overlay.set_show_icon(self.show_icon)
        self.overlay.set_reverse_order(self.reverse_order)
        self.overlay.set_limit_width(self.limit_width)
        self.overlay.set_icon_left(self.icon_left)
        self.overlay.set_icon_size(self.icon_size)
        self.overlay.set_icon_pad(self.icon_padding)
        self.overlay.set_padding(self.padding)
        self.overlay.set_border_radius(self.border_radius)
        if self.font:
            self.overlay.set_font(self.font)

    def save_config(self):
        """
        Save the current settings to the 'text' section of the config
        """
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        if not config.has_section("notification"):
            config.add_section("notification")

        config.set("notification", "rightalign", "%d" % (int(self.align_x)))
        config.set("notification", "topalign", "%d" % (self.align_y))
        config.set("notification", "monitor", self.monitor)
        config.set("notification", "enabled", "%d" % (int(self.enabled)))
        config.set("notification", "floating", "%s" % (int(self.floating)))
        config.set("notification", "floating_x", "%s" % (int(self.floating_x)))
        config.set("notification", "floating_y", "%s" % (int(self.floating_y)))
        config.set("notification", "floating_w", "%s" % (int(self.floating_w)))
        config.set("notification", "floating_h", "%s" % (int(self.floating_h)))
        config.set("notification", "bg_col", json.dumps(self.bg_col))
        config.set("notification", "fg_col", json.dumps(self.fg_col))
        config.set("notification", "text_time", "%s" % (int(self.text_time)))
        config.set("notification", "show_icon", "%s" %
                   (int(self.show_icon)))
        config.set("notification", "rev", "%s" %
                   (int(self.reverse_order)))
        config.set("notification", "limit_width", "%d" %
                   (int(self.limit_width)))
        config.set("notification", "icon_left", "%d" % (int(self.icon_left)))
        config.set("notification", "icon_padding", "%d" %
                   (int(self.icon_padding)))
        config.set("notification", "icon_size", "%d" % (int(self.icon_size)))
        config.set("notification", "padding", "%d" % (int(self.padding)))
        config.set("notification", "border_radius", "%d" %
                   (int(self.border_radius)))

        if self.font:
            config.set("notification", "font", self.font)

        with open(self.config_file, 'w') as file:
            config.write(file)

    def create_gui(self):
        """
        Prepare the gui
        """
        box = Gtk.Grid()

        # Enabled
        enabled_label = Gtk.Label.new(_("Enable"))
        enabled = Gtk.CheckButton.new()
        enabled.set_active(self.enabled)
        enabled.connect("toggled", self.change_enabled)

        # Enabled
        testing_label = Gtk.Label.new(_("Show test content"))
        testing = Gtk.CheckButton.new()
        testing.connect("toggled", self.change_testing)

        # Order
        reverse_label = Gtk.Label.new(_("Reverse Order"))
        reverse = Gtk.CheckButton.new()
        reverse.set_active(self.reverse_order)
        reverse.connect("toggled", self.change_reverse_order)

        # Popup timer
        text_time_label = Gtk.Label.new(_("Popup timer"))
        text_time_adjustment = Gtk.Adjustment.new(
            self.text_time, 8, 9000, 1, 1, 8)
        text_time = Gtk.SpinButton.new(text_time_adjustment, 0, 0)
        text_time.connect("value-changed", self.change_text_time)

        # Limit width
        limit_width_label = Gtk.Label.new(_("Limit popup width"))
        limit_width_adjustment = Gtk.Adjustment.new(
            self.limit_width, 100, 9000, 1, 1, 8)
        limit_width = Gtk.SpinButton.new(limit_width_adjustment, 0, 0)
        limit_width.connect("value-changed", self.change_limit_width)

        # Font chooser
        font_label = Gtk.Label.new(_("Font"))
        font = Gtk.FontButton()
        if self.font:
            font.set_font(self.font)
        font.connect("font-set", self.change_font)

        # Icon alignment
        align_icon_label = Gtk.Label.new(_("Icon position"))
        align_icon_store = Gtk.ListStore(str)
        align_icon_store.append([_("Left")])
        align_icon_store.append([_("Right")])
        align_icon = Gtk.ComboBox.new_with_model(align_icon_store)
        align_icon.set_active(not self.icon_left)
        align_icon.connect("changed", self.change_icon_left)
        renderer_text = Gtk.CellRendererText()
        align_icon.pack_start(renderer_text, True)
        align_icon.add_attribute(renderer_text, "text", 0)

        # Colours
        bg_col_label = Gtk.Label.new(_("Background colour"))
        bg_col = Gtk.ColorButton.new_with_rgba(
            Gdk.RGBA(self.bg_col[0], self.bg_col[1], self.bg_col[2], self.bg_col[3]))
        fg_col_label = Gtk.Label.new(_("Text colour"))
        fg_col = Gtk.ColorButton.new_with_rgba(
            Gdk.RGBA(self.fg_col[0], self.fg_col[1], self.fg_col[2], self.fg_col[3]))
        bg_col.set_use_alpha(True)
        fg_col.set_use_alpha(True)
        bg_col.connect("color-set", self.change_bg)
        fg_col.connect("color-set", self.change_fg)

        # Monitor & Alignment
        align_label = Gtk.Label.new(_("Overlay Location"))

        align_type_box = Gtk.HBox()
        align_type_edge = Gtk.RadioButton.new_with_label(
            None, _("Anchor to edge"))
        align_type_floating = Gtk.RadioButton.new_with_label_from_widget(
            align_type_edge, _("Floating"))
        if self.floating:
            align_type_floating.set_active(True)
        align_type_box.add(align_type_edge)
        align_type_box.add(align_type_floating)

        monitor_store = Gtk.ListStore(str)
        display = Gdk.Display.get_default()
        if "get_n_monitors" in dir(display):
            for i in range(0, display.get_n_monitors()):
                monitor_store.append([display.get_monitor(i).get_model()])
        monitor = Gtk.ComboBox.new_with_model(monitor_store)
        monitor.set_active(self.get_monitor_index(self.monitor))
        monitor.connect("changed", self.change_monitor)
        renderer_text = Gtk.CellRendererText()
        monitor.pack_start(renderer_text, True)
        monitor.add_attribute(renderer_text, "text", 0)

        align_x_store = Gtk.ListStore(str)
        align_x_store.append([_("Left")])
        align_x_store.append([_("Right")])
        align_x = Gtk.ComboBox.new_with_model(align_x_store)
        align_x.set_active(True if self.align_x else False)
        align_x.connect("changed", self.change_align_x)
        renderer_text = Gtk.CellRendererText()
        align_x.pack_start(renderer_text, True)
        align_x.add_attribute(renderer_text, "text", 0)

        align_y_store = Gtk.ListStore(str)
        align_y_store.append([_("Top")])
        align_y_store.append([_("Middle")])
        align_y_store.append([_("Bottom")])
        align_y = Gtk.ComboBox.new_with_model(align_y_store)
        align_y.set_active(self.align_y)
        align_y.connect("changed", self.change_align_y)
        renderer_text = Gtk.CellRendererText()
        align_y.pack_start(renderer_text, True)
        align_y.add_attribute(renderer_text, "text", 0)

        align_placement_button = Gtk.Button.new_with_label(_("Place Window"))

        align_type_edge.connect("toggled", self.change_align_type_edge)
        align_type_floating.connect("toggled", self.change_align_type_floating)
        align_placement_button.connect("pressed", self.change_placement)

        # Show Icons
        show_icon_label = Gtk.Label.new(_("Show Icon"))
        show_icon = Gtk.CheckButton.new()
        show_icon.set_active(self.show_icon)
        show_icon.connect("toggled", self.change_show_icon)

        # Icon Padding
        icon_padding_label = Gtk.Label.new(_("Icon padding"))
        icon_padding_adjustment = Gtk.Adjustment.new(
            self.icon_padding, 0, 150, 1, 1, 8)
        icon_padding = Gtk.SpinButton.new(icon_padding_adjustment, 0, 0)
        icon_padding.connect("value-changed", self.change_icon_pad)

        # Icon Size
        icon_size_label = Gtk.Label.new(_("Icon size"))
        icon_size_adjustment = Gtk.Adjustment.new(
            self.icon_size, 0, 128, 1, 1, 8)
        icon_size = Gtk.SpinButton.new(icon_size_adjustment, 0, 0)
        icon_size.connect("value-changed", self.change_icon_size)

        # Padding
        padding_label = Gtk.Label.new(_("Notification padding"))
        padding_adjustment = Gtk.Adjustment.new(
            self.padding, 0, 150, 1, 1, 8)
        padding = Gtk.SpinButton.new(padding_adjustment, 0, 0)
        padding.connect("value-changed", self.change_padding)

        # Border Radius
        border_radius_label = Gtk.Label.new(_("Border radius"))
        border_radius_adjustment = Gtk.Adjustment.new(
            self.border_radius, 0, 50, 1, 1, 8)
        border_radius = Gtk.SpinButton.new(border_radius_adjustment, 0, 0)
        border_radius.connect(
            "value-changed", self.change_border_radius)

        self.align_x_widget = align_x
        self.align_y_widget = align_y
        self.align_monitor_widget = monitor
        self.align_placement_widget = align_placement_button
        self.text_time_widget = text_time
        self.text_time_label_widget = text_time_label

        box.attach(enabled_label, 0, 0, 1, 1)
        box.attach(enabled, 1, 0, 1, 1)
        box.attach(reverse_label, 0, 1, 1, 1)
        box.attach(reverse, 1, 1, 1, 1)
        box.attach(text_time_label, 0, 3, 1, 1)
        box.attach(text_time, 1, 3, 1, 1)
        box.attach(limit_width_label, 0, 4, 1, 1)
        box.attach(limit_width, 1, 4, 1, 1)

        box.attach(font_label, 0, 6, 1, 1)
        box.attach(font, 1, 6, 1, 1)
        box.attach(fg_col_label, 0, 7, 1, 1)
        box.attach(fg_col, 1, 7, 1, 1)
        box.attach(bg_col_label, 0, 8, 1, 1)
        box.attach(bg_col, 1, 8, 1, 1)
        box.attach(align_label, 0, 9, 1, 5)
        #box.attach(align_type_box, 1, 8, 1, 1)
        box.attach(monitor, 1, 10, 1, 1)
        box.attach(align_x, 1, 11, 1, 1)
        box.attach(align_y, 1, 12, 1, 1)
        box.attach(align_placement_button, 1, 13, 1, 1)
        box.attach(show_icon_label, 0, 14, 1, 1)
        box.attach(show_icon, 1, 14, 1, 1)
        box.attach(align_icon_label, 0, 15, 1, 1)
        box.attach(align_icon, 1, 15, 1, 1)
        box.attach(icon_padding_label, 0, 16, 1, 1)
        box.attach(icon_padding, 1, 16, 1, 1)
        box.attach(icon_size_label, 0, 17, 1, 1)
        box.attach(icon_size, 1, 17, 1, 1)
        box.attach(padding_label, 0, 18, 1, 1)
        box.attach(padding, 1, 18, 1, 1)
        box.attach(border_radius_label, 0, 19, 1, 1)
        box.attach(border_radius, 1, 19, 1, 1)
        box.attach(testing_label, 0, 20, 1, 1)
        box.attach(testing, 1, 20, 1, 1)

        self.add(box)

    def change_padding(self, button):
        """
        Padding between notifications changed
        """
        self.overlay.set_padding(button.get_value())
        self.padding = button.get_value()

        self.save_config()

    def change_border_radius(self, button):
        """
        Border radius changed
        """
        self.overlay.set_border_radius(button.get_value())
        self.border_radius = button.get_value()

        self.save_config()

    def change_icon_size(self, button):
        """
        Icon size changed
        """
        self.overlay.set_icon_size(button.get_value())
        self.icon_size = button.get_value()

        self.save_config()

    def change_icon_pad(self, button):
        """
        Icon padding changed
        """
        self.overlay.set_icon_pad(button.get_value())
        self.icon_pad = button.get_value()

        self.save_config()

    def change_icon_left(self, button):
        """
        Icon alignment changed
        """
        self.overlay.set_icon_left(button.get_active() != 1)
        self.icon_left = (button.get_active() != 1)

        self.save_config()

    def change_font(self, button):
        """
        Font settings changed
        """
        font = button.get_font()
        self.overlay.set_font(font)

        self.font = font
        self.save_config()

    def change_text_time(self, button):
        """
        Popup style setting changed
        """
        self.overlay.set_text_time(button.get_value())

        self.text_time = button.get_value()
        self.save_config()

    def change_limit_width(self, button):
        """
        Popup width limiter
        """
        self.overlay.set_limit_width(button.get_value())
        self.limit_width = button.get_value()
        self.save_config()

    def change_bg(self, button):
        """
        Background colour changed
        """
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.overlay.set_bg(colour)

        self.bg_col = colour
        self.save_config()

    def change_fg(self, button):
        """
        Foreground colour changed
        """
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.overlay.set_fg(colour)

        self.fg_col = colour
        self.save_config()

    def change_show_icon(self, button):
        """
        Icon setting changed
        """
        self.overlay.set_show_icon(button.get_active())

        self.show_icon = button.get_active()
        self.save_config()

    def change_reverse_order(self, button):
        """
        Reverse Order changed
        """
        self.overlay.set_reverse_order(button.get_active())

        self.reverse_order = button.get_active()
        self.save_config()

    def change_testing(self, button):
        self.overlay.set_testing(button.get_active())

    def change_enabled(self, button):
        """
        Overlay active state toggled
        """
        self.overlay.set_enabled(button.get_active())
        self.enabled = button.get_active()
        self.save_config()



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Notification window for text"""
import logging
import time
import re
import cairo
import math
import gi
from .image_getter import get_surface, draw_img_to_rect, get_aspected_size, make_surface_from_raw
from .overlay import OverlayWindow
gi.require_version("Gtk", "3.0")
gi.require_version('PangoCairo', '1.0')
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Pango, PangoCairo  # nopep8

log = logging.getLogger(__name__)


class NotificationOverlayWindow(OverlayWindow):
    """Overlay window for notifications"""

    def __init__(self, discover, piggyback=None):
        OverlayWindow.__init__(self, discover, piggyback)
        self.text_spacing = 4
        self.content = []
        self.test_content = [{"icon": "https://cdn.discordapp.com/avatars/151774162749227008/75dd622dc9f5736f59f7ae1cfcf862a2.webp?size=128", "title": "Title1"},
                             {"title": "Title2", "body": "Body", "icon": None},
                             {"icon": "https://cdn.discordapp.com/avatars/151774162749227008/75dd622dc9f5736f59f7ae1cfcf862a2.webp?size=128", "title": "Title 3",
                                 "body": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."},
                             {"icon": None, "title": "Title 3", "body": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."},
                             {"icon": "https://cdn.discordapp.com/avatars/147077941317206016/0c0a10fa527c428e68fa527fe7a90bc5.webp?size=128", "title": "PM", "body": "Birdy test"}]
        self.text_font = None
        self.text_size = 13
        self.text_time = None
        self.show_icon = None
        self.pango_rect = Pango.Rectangle()
        self.pango_rect.width = self.text_size * Pango.SCALE
        self.pango_rect.height = self.text_size * Pango.SCALE

        self.connected = True
        self.bg_col = [0.0, 0.6, 0.0, 0.1]
        self.fg_col = [1.0, 1.0, 1.0, 1.0]
        self.icons = []
        self.reverse_order = False
        self.padding = 10
        self.border_radius = 5
        self.limit_width = 100
        self.testing = False
        self.icon_size = 64
        self.icon_pad = 16
        self.icon_left = True

        self.image_list = {}
        self.warned_filetypes = []
        self.set_title("Discover Text")
        self.redraw()

    def tick(self):
        # This doesn't really belong in overlay or settings
        now = time.time()
        newlist = []
        oldsize = len(self.content)
        # Iterate over and remove messages older than 30s
        for message in self.content:
            if message['time'] + self.text_time > now:
                newlist.append(message)
        self.content = newlist
        # If the list is different than before
        if oldsize != len(newlist):
            self.needsredraw = True

    def add_notification_message(self, data):
        noti = None
        data = data['data']
        message_id = data['message']['id']
        for message in self.content:
            if message['id'] == message_id:
                return
        if 'body' in data and 'title' in data:
            if 'icon_url' in data:
                noti = {"icon": data['icon_url'],
                        "title": data['title'],
                        "body": data['body'], "time": time.time(),
                        "id": message_id}
            else:
                noti = {"title": data['title'],
                        "body": data['body'], "time": time.time(),
                        "id": message_id}

        if noti:
            self.content.append(noti)
            self.needsredraw = True
            self.get_all_images()

    def set_padding(self, padding):
        """
        Set the padding between notifications
        """
        self.padding = padding
        self.needsredraw = True

    def set_border_radius(self, radius):
        """
        Set the radius of the border
        """
        self.border_radius = radius
        self.needsredraw = True

    def set_icon_size(self, size):
        """
        Set Icon size
        """
        self.icon_size = size
        self.image_list = {}
        self.get_all_images()

    def set_icon_pad(self, pad):
        """
        Set padding between icon and message
        """
        self.icon_pad = pad
        self.needsredraw = True

    def set_icon_left(self, left):
        self.icon_left = left
        self.needsredraw = True

    def set_text_time(self, timer):
        """
        Set the duration that a message will be visible for.
        """
        self.text_time = timer

    def set_limit_width(self, limit):
        """
        Set the word wrap limit in pixels
        """
        self.limit_width = limit
        self.needsredraw = True

    def get_all_images(self):
        the_list = self.content
        if self.testing:
            the_list = self.test_content
        for line in the_list:
            icon = line["icon"]

            if icon and icon not in self.image_list:
                get_surface(self.recv_icon, icon, icon,
                            self.icon_size)

    def recv_icon(self, identifier, pix):
        """
        Called when image_getter has downloaded an image
        """
        self.image_list[identifier] = pix
        self.needsredraw = True

    def set_fg(self, fg_col):
        """
        Set default text colour
        """
        self.fg_col = fg_col
        self.needsredraw = True

    def set_bg(self, bg_col):
        """
        Set background colour
        """
        self.bg_col = bg_col
        self.needsredraw = True

    def set_show_icon(self, icon):
        """
        Set if icons should be shown inline
        """
        self.show_icon = icon
        self.needsredraw = True
        self.get_all_images()

    def set_reverse_order(self, rev):
        self.reverse_order = rev
        self.needsredraw = True

    def set_font(self, font):
        """
        Set font used to render text
        """
        self.text_font = font

        self.pango_rect = Pango.Rectangle()
        font = Pango.FontDescription(self.text_font)
        self.pango_rect.width = font.get_size() * Pango.SCALE
        self.pango_rect.height = font.get_size() * Pango.SCALE
        self.needsredraw = True

    def recv_attach(self, identifier, pix):
        """
        Called when an image has been downloaded by image_getter
        """
        self.icons[identifier] = pix
        self.needsredraw = True

    def calc_all_height(self):
        h = 0
        my_list = self.content
        if self.testing:
            my_list = self.test_content
        for line in my_list:
            h += self.calc_height(line)
        if h > 0:
            h -= self.padding  # Remove one unneeded padding
        return h

    def calc_height(self, line):
        icon_width = 0
        icon_pad = 0
        icon = line['icon']
        if self.show_icon and icon and icon in self.image_list and self.image_list[icon]:
            icon_width = self.icon_size
            icon_pad = self.icon_pad
        message = ""
        if 'body' in line and len(line['body']) > 0:
            m_no_body = "<span>%s</span>\n%s"
            message = m_no_body % (self.sanitize_string(line["title"]),
                                   self.sanitize_string(line['body']))
        else:
            m_with_body = "<span>%s</span>"
            message = m_with_body % (self.sanitize_string(line["title"]))
        layout = self.create_pango_layout(message)
        layout.set_auto_dir(True)
        layout.set_markup(message, -1)
        attr = layout.get_attributes()
        width = self.limit_width if self.width > self.limit_width else self.width
        layout.set_width((Pango.SCALE * (width -
                         (self.border_radius * 4 + icon_width + icon_pad))))
        layout.set_spacing(Pango.SCALE * 3)
        if self.text_font:
            font = Pango.FontDescription(self.text_font)
            layout.set_font_description(font)
        text_width, text_height = layout.get_pixel_size()
        if text_height < icon_width:
            text_height = icon_width
        return text_height + (self.border_radius*4) + self.padding

    def overlay_draw(self, _w, context, data=None):
        """
        Draw the overlay
        """
        if self.piggyback:
            self.piggyback.overlay_draw(w, context, data)
        if not self.enabled:
            return
        self.context = context
        (width, height) = self.get_size()
        if not self.piggyback_parent:
            context.set_antialias(cairo.ANTIALIAS_GOOD)

            # Make background transparent
            context.set_source_rgba(0.0, 0.0, 0.0, 0.0)
            context.set_operator(cairo.OPERATOR_SOURCE)
            context.paint()

        context.save()
        if self.is_wayland or self.piggyback_parent or self.discover.steamos:
            # Special case!
            # The window is full-screen regardless of what the user has selected.
            # We need to set a clip and a transform to imitate original behaviour
            # Used in wlroots & gamescope
            width = self.width
            height = self.height
            context.translate(self.pos_x, self.pos_y)
            context.rectangle(0, 0, width, height)
            context.clip()

        current_y = height
        if self.align_vert == 0:
            current_y = 0
        if self.align_vert == 1:  # Center. Oh god why
            current_y = (height/2.0) - (self.calc_all_height() / 2.0)
        tnow = time.time()
        if self.testing:
            the_list = self.test_content
        else:
            the_list = self.content
        if self.reverse_order:
            the_list = reversed(the_list)
        for line in the_list:
            col = "#fff"
            if 'body' in line and len(line['body']) > 0:
                m_no_body = "<span foreground='%s'>%s</span>\n%s"
                message = m_no_body % (self.sanitize_string(col),
                                       self.sanitize_string(line["title"]),
                                       self.sanitize_string(line['body']))
            else:
                m_with_body = "<span foreground='%s'>%s</span>"
                message = m_with_body % (self.sanitize_string(col),
                                         self.sanitize_string(line["title"]))

            icon = None
            # If we've got an embedded image
            if "icon_surface" in line and line["icon_surface"]:
                icon = line["icon_surface"]
            # If we're given an icon name, it's in the list of icons, and it's not none
            elif line["icon"] and line["icon"] in self.image_list and self.image_list[line["icon"]]:
                icon = self.image_list[line["icon"]]

            current_y = self.draw_text(current_y, message, icon)
            if current_y <= 0:
                # We've done enough
                break
        context.restore()
        self.context = None

    def draw_text(self, pos_y, text, icon):
        """
        Draw a text message, returning the Y position of the next message
        """

        icon_width = self.icon_size
        icon_pad = self.icon_pad
        if not self.show_icon:
            icon = None
        if not icon:
            icon_pad = 0
            icon_width = 0

        layout = self.create_pango_layout(text)
        layout.set_auto_dir(True)
        layout.set_markup(text, -1)
        attr = layout.get_attributes()

        width = self.limit_width if self.width > self.limit_width else self.width
        layout.set_width((Pango.SCALE * (width -
                         (self.border_radius * 4 + icon_width + icon_pad))))
        layout.set_spacing(Pango.SCALE * 3)
        if self.text_font:
            font = Pango.FontDescription(self.text_font)
            layout.set_font_description(font)
        text_width, text_height = layout.get_pixel_size()
        self.col(self.bg_col)
        top = 0
        if self.align_vert == 2:  # Bottom align
            top = pos_y - (text_height + self.border_radius * 4)
        else:  # Top align
            top = pos_y
        if text_height < icon_width:
            text_height = icon_width
        shape_height = text_height + self.border_radius * 4
        shape_width = text_width + self.border_radius*4 + icon_width + icon_pad

        left = 0
        if self.align_right:
            left = self.width - shape_width

        self.context.save()
        # Draw Background
        self.context.translate(left, top)
        # self.context.rectangle(self.border_radius, 0,
        #                       shape_width - (self.border_radius*2), shape_height)
        # self.context.fill()
        # self.context.rectangle(0, self.border_radius,
        #                       shape_width, shape_height - (self.border_radius * 2))

        #self.context.arc(0.7, 0.3, 0.035, 1.25 * math.pi, 2.25 * math.pi)
        #self.context.arc(0.3, 0.7, 0.035, .25 * math.pi, 1.25 * math.pi)
        if self.border_radius == 0:

            self.context.move_to(0.0, 0.0)
            self.context.line_to(shape_width, 0.0)
            self.context.line_to(shape_width, shape_height)
            self.context.line_to(0.0, shape_height)
            self.context.close_path()
            self.context.fill()
        else:
            # Edge top
            self.context.move_to(self.border_radius, 0.0)
            self.context.line_to(shape_width - self.border_radius, 0.0)

            # Arc topright
            self.context.arc(shape_width - self.border_radius, self.border_radius,
                             self.border_radius, 1.5 * math.pi, 2 * math.pi)

            # Edge right
            self.context.line_to(
                shape_width, shape_height - self.border_radius)

            # Arc bottomright
            self.context.arc(shape_width - self.border_radius, shape_height - self.border_radius,
                             self.border_radius, 0.0, 0.5 * math.pi)

            # Edge bottom
            self.context.line_to(self.border_radius, shape_height)

            # Arch bottomleft
            self.context.arc(self.border_radius, shape_height -
                             self.border_radius, self.border_radius, 0.5 * math.pi, math.pi)

            # Edge left
            self.context.line_to(0.0, self.border_radius)

            # Arc topleft
            self.context.arc(self.border_radius, self.border_radius,
                             self.border_radius, math.pi, 1.5 * math.pi)

            # End
            self.context.close_path()
            self.context.fill()

        self.context.set_operator(cairo.OPERATOR_OVER)
        # Draw Image
        if icon:
            self.context.save()
            if self.icon_left:
                self.context.translate(
                    self.border_radius*2, self.border_radius*2)
                draw_img_to_rect(icon, self.context, 0, 0,
                                 icon_width, icon_width)
            else:
                self.context.translate(
                    self.border_radius*2 + text_width + icon_pad, self.border_radius*2)
                draw_img_to_rect(icon, self.context, 0, 0,
                                 icon_width, icon_width)
            self.context.restore()

        self.col(self.fg_col)

        if self.icon_left:
            self.context.translate(
                self.border_radius*2 + icon_width + icon_pad, self.border_radius*2)
            PangoCairo.context_set_shape_renderer(
                self.get_pango_context(), self.render_custom, None)

            text = layout.get_text()
            count = 0

            layout.set_attributes(attr)

            PangoCairo.show_layout(self.context, layout)
        else:
            self.context.translate(self.border_radius*2, self.border_radius*2)
            PangoCairo.context_set_shape_renderer(
                self.get_pango_context(), self.render_custom, None)

            text = layout.get_text()
            count = 0

            layout.set_attributes(attr)

            PangoCairo.show_layout(self.context, layout)

        self.context.restore()
        next_y = 0
        if self.align_vert == 2:
            next_y = pos_y - (shape_height + self.padding)
        else:
            next_y = pos_y + shape_height + self.padding
        return next_y

    def render_custom(self, ctx, shape, path, _data):
        """
        Draw an inline image as a custom emoticon
        """
        if shape.data >= len(self.image_list):
            log.warning(f"{shape.data} >= {len(self.image_list)}")
            return
        # key is the url to the image
        key = self.image_list[shape.data]
        if key not in self.attachment:
            get_surface(self.recv_attach,
                        key,
                        key, None)
            return
        pix = self.attachment[key]
        (pos_x, pos_y) = ctx.get_current_point()
        draw_img_to_rect(pix, ctx, pos_x, pos_y - self.text_size, self.text_size,
                         self.text_size, path=path)
        return True

    def sanitize_string(self, string):
        """
        Sanitize a text message so that it doesn't intefere with Pango's XML format
        """
        string = string.replace("&", "&amp;")
        string = string.replace("<", "&lt;")
        string = string .replace(">", "&gt;")
        string = string.replace("'", "&#39;")
        string = string.replace("\"", "&#34;")
        return string

    def set_testing(self, testing):
        self.testing = testing
        self.needsredraw = True
        self.get_all_images()



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Core Settings Tab"""
import gettext
from configparser import ConfigParser
import gi
import pkg_resources
from .settings import SettingsWindow
from .autostart import Autostart
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk  # nopep8

t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'))
_ = t.gettext


class GeneralSettingsWindow(SettingsWindow):
    """Core Settings Tab"""

    def __init__(self, discover):
        SettingsWindow.__init__(self, discover)
        self.discover = discover
        self.xshape = None
        self.show_sys_tray_icon = None
        self.show_task = None
        self.set_size_request(400, 200)
        self.connect("destroy", self.close_window)
        self.connect("delete-event", self.close_window)
        self.init_config()
        self.autostart_helper = Autostart("discover_overlay")
        self.placement_window = None

        self.create_gui()

    def read_config(self):
        """
        Read in the 'general' section of config and set overlays
        """
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        self.xshape = config.getboolean("general", "xshape", fallback=False)
        self.show_sys_tray_icon = config.getboolean(
            "general", "showsystray", fallback=True)
        self.show_task = config.getboolean(
            "general", "showtask", fallback=False)

        # Pass all of our config over to the overlay
        self.discover.set_force_xshape(self.xshape)
        self.discover.set_sys_tray_icon_visible(self.show_sys_tray_icon)

    def save_config(self):
        """
        Save the 'general' section of config
        """
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        if not config.has_section("general"):
            config.add_section("general")

        config.set("general", "xshape", "%d" % (int(self.xshape)))
        config.set("general", "showsystray",
                   "yes" if self.show_sys_tray_icon else "no")
        config.set("general", "showtask", "yes" if self.show_task else "no")

        with open(self.config_file, 'w') as file:
            config.write(file)

    def create_gui(self):
        """
        Prepare the GUI
        """
        box = Gtk.Grid()

        # Auto start
        autostart_label = Gtk.Label.new(_("Autostart on boot"))
        autostart = Gtk.CheckButton.new()
        autostart.set_active(self.autostart_helper.is_auto())
        autostart.connect("toggled", self.change_autostart)

        # Force XShape
        xshape_label = Gtk.Label.new(_("Force XShape"))
        xshape = Gtk.CheckButton.new()
        xshape.set_active(self.xshape)
        xshape.connect("toggled", self.change_xshape)

        # Show sys tray
        show_sys_tray_icon_label = Gtk.Label.new(_("Show tray icon"))
        show_sys_tray_icon = Gtk.CheckButton.new()
        show_sys_tray_icon.set_active(self.show_sys_tray_icon)
        show_sys_tray_icon.connect("toggled", self.change_show_sys_tray_icon)

        # Show taskbar
        show_task_label = Gtk.Label.new(_("Show on taskbar"))
        show_task = Gtk.CheckButton.new()
        show_task.set_active(self.show_task)
        show_task.connect("toggled", self.change_show_task)

        box.attach(autostart_label, 0, 0, 1, 1)
        box.attach(autostart, 1, 0, 1, 1)
        box.attach(xshape_label, 0, 1, 1, 1)
        box.attach(xshape, 1, 1, 1, 1)
        box.attach(show_sys_tray_icon_label, 0, 2, 1, 1)
        box.attach(show_sys_tray_icon, 1, 2, 1, 1)
        box.attach(show_task_label, 0, 3, 1, 1)
        box.attach(show_task, 1, 3, 1, 1)

        self.add(box)

    def change_autostart(self, button):
        """
        Autostart setting changed
        """
        autostart = button.get_active()
        self.autostart_helper.set_autostart(autostart)

    def change_xshape(self, button):
        """
        XShape setting changed
        """
        self.discover.set_force_xshape(button.get_active())
        self.xshape = button.get_active()
        self.save_config()

    def change_show_sys_tray_icon(self, button):
        """
        Show tray icon setting changed
        """
        self.discover.set_sys_tray_icon_visible(button.get_active())
        self.show_sys_tray_icon = button.get_active()
        self.save_config()

    def change_show_task(self, button):
        """
        Show in task bar changed
        """
        self.discover.set_show_task(button.get_active())
        self.show_task = button.get_active()
        self.save_config()
